import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import java.util.Arrays; 
import java.util.HashSet; 
import java.util.Iterator; 
import java.util.Collections; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class omnigram extends PApplet {

/**
* Omnigram Explorer
* An Interactive Data Exploration Tool
*
* Tim Taylor
* Monash University
* tim@tim-taylor.com
*
* version: 1.1
* date: 14 March 2015
*
* Notes on exporting to platform-specific binaries:
* - Export to Linux works best NOT as a full screen mode app [and with size less than full screen in setup()]
* - Export to Windows works best as a full screen mode app [and with size = full screen in setup()]
* - Export to MacOS: for current version of Processing (2.2.1), only works when exporting from a Mac
*
*/

String modelLoaderFile = "";
int globalZoom = 100;
int nodeZoom = 100;
Model model;


public void setup() {
  size((displayWidth*80)/100, (displayHeight*80)/100); // best for Linux
  //size(displayWidth, displayHeight); // best for Windows
  if (frame != null) {
    frame.setResizable(true);
  }
  smooth();
  noStroke();
  noLoop();
  selectInput("Select a model definition XML file", "modelFileSelected");
}


public void modelFileSelected(File file) {
  // Callback function for the file selection dialog
  if (file == null) {
    exit();
  } else {
    modelLoaderFile = file.getAbsolutePath();
    model = new Model(modelLoaderFile);
    model.m_interactionMode = InteractionMode.SingleNodeBrushing;
    loop();
  }
}


public void draw() {
  if (model != null) {
    model.draw(globalZoom, nodeZoom);
  }
}


public void mousePressed() {
  model.mousePressed();
}


public void mouseReleased() {  
  model.mouseReleased();
}


public void mouseDragged() {
  model.mouseDragged();
}


public void keyPressed() {
  if (key == CODED) {
    switch (keyCode) {
      case UP:
        model.showSamplesAutoSpeedUp();
        break;
      case DOWN:
        model.showSamplesAutoSlowDown();
        break;
      case LEFT:
        model.showSamplesStepBackward();
        break;
      case RIGHT:
        model.showSamplesStepForward();
        break;
      default:
        break;
    }
  }
  else {
    switch (key) {
      case '+': {
        if (globalZoom < 1000) { 
          globalZoom++;
          redraw();
        }
        break;
      }
      case '-': {
        if (globalZoom > 1) {
          globalZoom--;
          redraw();
        }
        break;
      }
      case '1':
        model.setInteractionMode(InteractionMode.SingleNodeBrushing);
        break;
      case '2':
        model.setInteractionMode(InteractionMode.MultiNodeBrushing);
        break;
      case '3':
        model.setInteractionMode(InteractionMode.ColourBins);
        break;
      case '4':
        model.setInteractionMode(InteractionMode.ShowSamples);
        break;        
      case 'B':
      case 'b':
        model.requestBrushLinkDelete();
        break;
      case 'C':
      case 'c':
        model.requestCausalLinkDelete();
        break;
      case 'D':
      case 'd':
        model.requestToggleCausalLinkDir();
        break;
      case 'H':
      case 'h':  
        model.toggleHelpScreen();
        break;
      case 'L':
      case 'l':
        model.requestBrushLinkCreate();
        break;
      case 'M':
      case 'm':
        model.toggleMeanMedian();
        break;
      case 'R':
      case 'r':
        model.initialiseCausalLinks();
      case 'S':
      case 's':
        model.toggleSSDisplayMode();
        break;
      case 'Z':
      case 'z':
        model.showSamplesDecrementNumSamples();
        break;
      case 'X':
      case 'x':
        model.showSamplesIncrementNumSamples();
        break;
      default:
        break;
    }
  }
}
  

public class BrushLink {
  
  Model m_model;
  Node  m_node1;
  Node  m_node2;
  float m_strength;
  
  PVector m_chPos; // position of the link's control handle
  int m_chSize;    // size of the control handle
  
  int m_strokeCol;
  int m_chHighlightCol;
  int m_chBackgroundCol;

  
  BrushLink(Model model, Node n1, Node n2, float strength) {
    m_model = model;
    m_node1 = n1;
    m_node2 = n2;
    m_strength = strength;
    m_chPos = new PVector();
    m_chSize = 20;
    m_strokeCol = 0xff000000;
    m_chHighlightCol = 0xffFF0000;
    m_chBackgroundCol = 0xffFFFFFF;
  }
  
  
  public void draw() {
    
    if (!enabled()) {
      return;
    }    
    
    pushMatrix();
    pushStyle();
    
    int n1x = m_node1.getCentreX();
    int n1y = m_node1.getCentreY();
    int n2x = m_node2.getCentreX();
    int n2y = m_node2.getCentreY();
    
    PVector v1 = new PVector(n1x, n1y);
    PVector v2 = new PVector(n2x, n2y);
    PVector v1v2 = PVector.sub(v2, v1);
    m_chPos = PVector.add(v1, PVector.mult(v1v2, 0.5f));    
    
    scale(((float)m_model.m_globalZoom)/100.0f);    
    
    // draw line
    stroke(m_strokeCol);
    strokeWeight(2);
    line(v1.x, v1.y, v2.x, v2.y);
    
    // draw control handle
    rectMode(CENTER);
    
    strokeWeight(1);
    fill(m_chBackgroundCol);
    rect(m_chPos.x, m_chPos.y, m_chSize, m_chSize);
    
    //stroke(m_strokeCol);
    noStroke();
    fill(m_strokeCol);
    rect(m_chPos.x, m_chPos.y - (int)(((m_strength * (float)m_chSize) / 4.0f)), 
         m_chSize, (int)((m_strength * (float)m_chSize) / 2.0f));
    
    noFill();
    strokeWeight(2);
    stroke( mouseOver() ? m_chHighlightCol : m_strokeCol );
    rect(m_chPos.x, m_chPos.y, m_chSize, m_chSize);    
    
    popStyle();
    popMatrix();
  }
  
  
  public boolean mouseOver() {
    // is the mouse pointer currently over this link's control handle
    int r = m_chSize / 2;
    int mx = m_node1.scaledMouseX();
    int my = m_node1.scaledMouseY();
    return (mx >= (m_chPos.x - r) &&
            mx <  (m_chPos.x + r) &&
            my >= (m_chPos.y - r) &&
            my <  (m_chPos.y + r));    
  }
  
  
  public void mousePressed() {
    if (mouseOver()) {
      if (m_node1.scaledMouseY() <= m_chPos.y) {
        m_strength += 0.2f;
      }
      else {
        m_strength -= 0.2f;
      }
      m_strength = constrain(m_strength, -1.0f, 1.0f);
    }
  }
  
  
  public boolean enabled() {
    return !(m_node1.minimised() || m_node2.minimised());
  }
  
}
public class CausalLink {
  
  Model m_model;
  Node  m_nodeFrom;
  Node  m_nodeTo;
  
  int     m_chSizeBy2;  // defines the size of the control handle
  PVector m_chPos;      // current position of the link's control handle
  float   m_chAngle;    // current angle of the link's control handle
  
  int m_strokeCol;
  int m_chHighlightCol;
  int m_chBackgroundCol;

  
  CausalLink(Model model, Node nodeFrom, Node nodeTo) {
    m_model = model;
    m_nodeFrom = nodeFrom;
    m_nodeTo = nodeTo;
    m_chPos = new PVector();
    m_chSizeBy2 = 10;
    m_chAngle = 0;
    m_strokeCol = 0xff000000;
    m_chHighlightCol = 0xffFF0000;
    m_chBackgroundCol = 0xffFFFFFF;
  }
  
  
  public void draw() {
    
    if (!enabled()) {
      return;
    }
    
    pushMatrix();
    pushStyle();
    
    int nFx = m_nodeFrom.getCentreX();
    int nFy = m_nodeFrom.getCentreY();
    int nTx = m_nodeTo.getCentreX();
    int nTy = m_nodeTo.getCentreY();
    
    PVector vF = new PVector(nFx, nFy);
    PVector vT = new PVector(nTx, nTy);
    PVector vFvT = PVector.sub(vT, vF);
    m_chPos = PVector.add(vF, PVector.mult(vFvT, 0.5f));    
    
    scale(((float)m_model.m_globalZoom)/100.0f);    
    
    // draw line
    stroke(m_strokeCol);
    strokeWeight(2);
    line(vF.x, vF.y, vT.x, vT.y);
       
    // draw control handle (an arrow pointing from nodeFrom to nodeTo)
    m_chAngle = atan2(vFvT.x, vFvT.y);
    translate(m_chPos.x, m_chPos.y);
    rotate(PI-m_chAngle);
    fill(m_chBackgroundCol);
    triangle(0,-m_chSizeBy2, m_chSizeBy2,m_chSizeBy2, -m_chSizeBy2,m_chSizeBy2);
    // draw outline of handle according to whether mouse is over it
    noFill();
    strokeWeight(2);
    stroke( mouseOver() ? m_chHighlightCol : m_strokeCol );
    triangle(0,-m_chSizeBy2, m_chSizeBy2,m_chSizeBy2, -m_chSizeBy2,m_chSizeBy2);    
    
    popStyle();
    popMatrix();
  }
  
  
  public boolean mouseOver() {
    // is the mouse pointer currently over this link's control handle
    int mx = m_nodeFrom.scaledMouseX();
    int my = m_nodeFrom.scaledMouseY();
    return (dist(mx,my, m_chPos.x,m_chPos.y) <= m_chSizeBy2);   
  }
  
  
  public void toggleDirection() {
    // reverse the direction of the to/from nodes 
    Node tmp = m_nodeFrom;
    m_nodeFrom = m_nodeTo;
    m_nodeTo = tmp;
  }
  
  
  public boolean enabled() {
    return !(m_nodeTo.minimised() || m_nodeFrom.minimised());
  }  
  
}
public class ContinuousNode extends Node {

  // Range selector info
  float m_rangeMin;      // min value selectable
  float m_rangeMax;      // max value selectable
  float m_rangeLow;      // current low point selected
  float m_rangeHigh;     // current high point selected
  
 
 
  ContinuousNode(Model model, int id, String name, int filecol, int datacol, float min, float max, int role, ArrayList<Integer> parentIDs) {
    super(model, id, name, filecol, datacol, role, parentIDs);
    m_rangeMin = m_rangeLow = min;
    m_rangeMax = m_rangeHigh = max;
  }  
  
  
  public int getSelectedRange() {
    return (int)(m_rangeHigh - m_rangeLow);
  }


  public int getFullRange() {
    return (int)(m_rangeMax - m_rangeMin);
  }
  
  
  public void initialiseHistogram() {
    m_hgNumBins = m_sMaxBins;
    
    /*
    println("Histogram boundary values for "+m_name);
    for (int i=0; i<m_hgNumBins; i++) {
      println(getHistogramBinLowVal(i));
    }
    */   
    
    initialiseHistogramCommon(); 
  }
  
  
  public int getHistogramIndex(Number num) {
    // Return the index in node.m_hgBins corresponding to the data value passed in (i.e. which
    // bin does that value belong to). Returns -1 if num is outside of the defined range of the Node
    float fNum = num.floatValue();
    if ((fNum < (m_rangeMin - 0.00001f)) || (fNum > (m_rangeMax + 0.00001f))) {
      return -1;
    }
    else {
      int idx = (int)(((fNum - m_rangeMin) * m_hgNumBins) / (m_rangeMax - m_rangeMin));
      if (idx==m_hgNumBins) idx--; // not very elegant, but it works
      return idx;
    }
  }
  
  
  public Number getHistogramBinLowVal(int bin) {
    return new Float((((bin*(m_rangeMax - m_rangeMin))/m_hgNumBins) + m_rangeMin));
  }
  
  
  public Number getHistogramBinHighVal(int bin) {
    Float high = new Float((Float)getHistogramBinLowVal(bin+1));
    return (high);
  }  

}

public class DiscreteNode extends Node {
  
  // Range selector info
  int m_rangeMin;      // min value selectable
  int m_rangeMax;      // max value selectable
  int m_rangeLow;      // current low point selected
  int m_rangeHigh;     // current high point selected  
  
  
  DiscreteNode(Model model, int id, String name, int filecol, int datacol, int min, int max, int role, ArrayList<Integer> parentIDs) {
    super(model, id, name, filecol, datacol, role, parentIDs);
    m_rangeMin = m_rangeLow = min;
    m_rangeMax = m_rangeHigh = max;
  }

  
  public int getSelectedRange() {
    return m_rangeHigh - m_rangeLow + 1;
  }


  public int getFullRange() {
    return m_rangeMax - m_rangeMin + 1;
  }
  
  
  public void initialiseHistogram() {
    int dataRange = getFullRange();
    m_hgNumBins = (dataRange > m_sMaxBins) ? m_sMaxBins : dataRange;
    
    /*
    println("Histogram boundary values for "+m_name);
    for (int i=0; i<m_hgNumBins; i++) {
      println(getHistogramBinLowVal(i));
    }
    */
    
    initialiseHistogramCommon();
  }
  
  
  public int getHistogramIndex(Number num) {
    // Return the index in node.m_hgBins corresponding to the data value passed in (i.e. which
    // bin does that value belong to). Returns -1 if num is outside of the defined range of the Node
    
    int iNum = num.intValue();
    if (iNum < m_rangeMin || iNum > m_rangeMax) {
      return -1;
    }
    else {
      return (((iNum - m_rangeMin) * m_hgNumBins) / (m_rangeMax - m_rangeMin + 1));
    }
  }
  
  
  public Number getHistogramBinLowVal(int bin) {
    return new Integer((((bin*(m_rangeMax - m_rangeMin + 1))/m_hgNumBins) + m_rangeMin));
  }
  
  
  public Number getHistogramBinHighVal(int bin) {
    Integer high = new Integer((Integer)getHistogramBinLowVal(bin+1));
    return (high - 1);
  }

}

class HistogramBin {
  
  // relationship to owning Node object
  Node m_node; // reference to associated node
  int m_idx;   // 0-based index to this bin, as used in m_node.m_hgBins
  
  // info on the sampleIDs held in this bin
  ArrayList<Integer> m_sampleIDs;        // IDs (zero-based array indices in m_node.m_model.m_data) of data samples in this bin
  ArrayList<Integer> m_brushedSampleIDs; // IDs of currently brushed (numMisses==0) samples in this bin
  int m_numSamples;                      // a convenience variable (DERIVED = m_sampleIDs.length())
  
  // color definitions
  int m_strokeColor = 0xFF000000;
  int m_focalInRangeSelFillColor = 0xFF880000;
  int m_focalOutsideRangeSelFillColor = 0xFFFFFFFF;
  int m_nonFocalNonBrushedFillColor = 0xFFFFFFFF;
  int[] m_brushColors;
  
  int[] m_numBrushedSamples; // records number of brush samples in bin, where index is number of 
                             // near misses. So idx 0 is samples that exactly match selected range
                             // idx 1 is samples with one near miss, etc.

  int m_x; // x pos of bottom-left corner of bin, relative to left edge of histogram window
  int m_y; // y pos of bottom-left corner of bin, relative to top edge of histogram window
  int m_w; // width of bin
  int m_h; // height of bin
  int m_referenceH; // reference height for the bin, used when rescaling
  
  HistogramBin(Node node, int idx, int numSamples, ArrayList<Integer> sampleIDs, int x, int y) {
    m_node = node;
    m_idx = idx;
    m_sampleIDs = sampleIDs;
    m_numSamples = sampleIDs.size();
    m_brushedSampleIDs = new ArrayList<Integer>();
    m_brushColors = new int[node.m_hgNumBrushes];
    m_numBrushedSamples = new int[node.m_hgNumBrushes];

    for (int i=0; i<node.m_hgNumBrushes; i++) {
      m_numBrushedSamples[i] = 0;
      switch (i) {
        case 0:  m_brushColors[i] = 0xFF000088; break; // brush color for exact match
        case 1:  m_brushColors[i] = 0xFFB0E0B0; break; // brush color for one misses
        case 2:  m_brushColors[i] = 0xFFE0B0B0; break; // brush color for two misses
        default: m_brushColors[i] = 0xFFFFFFFF; break; // brush color for three or more misses
      }
    }
    
    m_x = x;
    m_y = y;
    setBinDimensions(); // set m_w and m_h
  }
  
  
  public void setBinDimensions() {
    // calculate the bin's width and height
    
    int dh = m_node.m_hgDefaultMaxBinH;
    int dw = m_node.m_hgDefaultBinW;
    int numSamplesAll = m_node.m_model.m_data.size();
    float nodeSF = m_node.m_model.m_nodeBinScaleFactor;

    m_w = dw;
    m_h = (int)((nodeSF * (float)(dh * m_numSamples))/((float)numSamplesAll));
    m_referenceH = m_h;
  }
  
  
  public void setX(int x) {
    m_x = x;
  }
  
  
  public void setY(int y) {
    m_y = y;
  }
  
  
  public void scaleH(float sf, boolean resetReferenceH) {
    //m_h = (int)(((float)m_h) * sf);
    m_h = (int)(((float)m_referenceH) * sf);
    if (resetReferenceH) {
      m_referenceH = m_h;
    }
    //println("bin.scaleH: sf="+sf+", m_h="+m_h+", m_refH="+m_referenceH);
  }
  
  
  public int getH() {
    return m_h;
  }
  
  
  public int getW() {
    return m_w;
  }
  
  
  public int getLX() {
    // return the x pos of the left hand side of the bin
    return m_x;
  }
  
  
  public int getRX() {
    // return the x pos of the right hand side of the bin
    return m_x + m_w;
  }
  
  
  public void draw() {
    // draw the bin
    
    pushStyle();   

    if (m_h > 0) {
      stroke(m_strokeColor);
      
      switch (m_node.m_model.m_interactionMode) {
        case SingleNodeBrushing: 
        case MultiNodeBrushing:
        case ShowSamples: {
          if (m_node.hasFocus()) {
            // this is the focal node for single-node brushing
            if (inSelectedRange()) {
              fill(m_focalInRangeSelFillColor);
            }
            else {
              fill(m_focalOutsideRangeSelFillColor);
            }
            rect(m_x, m_y-m_h, m_w, m_h);
          }
          else {
            // this is a non-focal node for single-node brushing
            boolean singleNodeBrushing = (m_node.m_model.m_interactionMode != InteractionMode.MultiNodeBrushing);
            drawBrushedBin(singleNodeBrushing ? 1 : m_node.m_hgNumBrushes);
          }
          
          // in ShowSamples mode, draw the individual samples
          if (m_node.m_model.m_interactionMode == InteractionMode.ShowSamples) {
            drawIndividualSamples();
          }
          /*
          else if (m_node.m_model.m_interactionMode == InteractionMode.ColourBins) {
            drawColourBins();
          }
          */
          
          break;
        }
        case ColourBins: {      
          drawColourBins();
          break;
        }
        case Unassigned:
        default: {
          println("Unexpected case found in HistogramBin::drawFullAutoHeightAdjust!");
          exit();
        }
      }
    }
    
    // draw the base under each bin
    stroke(m_node.m_hgBaseColor);
    strokeWeight(m_node.m_hgBaseStrokeWeight);
    int baseY = m_y + (m_node.m_hgFootH / 2);
    line(m_x, baseY, m_x+m_w, baseY);
    
    popStyle();
  }
  
  
  public void drawIndividualSamples() {
    // helper method for drawing
    // draws individual samples as circles in the correct colour for ShowSamples mode
    
    pushStyle();
    
    if (!m_node.hasFocus()) {
      // in ShowSamples mode, partially fade out bins in non-focal nodes
      fill(255, 220);
      rect(m_x, m_y-m_h, m_w, m_h);
    }
    
    colorMode(HSB);
    stroke(0);
    int numInBin = 0;
    ArrayList<Integer> samples = m_node.m_model.m_ssSamplesToDisplay;     // a list of all sampleIDs currently selected
    int dIdx = m_node.m_model.m_ssDisplayIdx;          // index of first sample in m_ssSamplesToDisplay to be displayed
    int numSamplesToDisplay = m_node.m_model.m_ssMaxSamplesToDisplay;
                
    if (!samples.isEmpty()) {
      
      // For model.m_ssDisplayMode==1|2, we need to first work out the color of each sample to
      // be displayed in this bin, then sort the colors so that we display contiguous blocks
      // of each color.
      ArrayList<Integer> sortedHues = new ArrayList<Integer>();
      if (m_node.m_model.m_ssDisplayMode == 1 || m_node.m_model.m_ssDisplayMode == 2) {
        for (int i = 0; i < numSamplesToDisplay; i++) {
          int sID = samples.get( (dIdx+i) % (samples.size()) );
          int sIdx = m_sampleIDs.indexOf(sID);
          if (sIdx >= 0) {                             // this sample appears in this bin, so record its color               
            sortedHues.add(m_node.m_model.getSampleHue(dIdx+i));
          }
        }
        Collections.sort(sortedHues);
      }
      
      // Now loop through each sample to be displayed, and check whether it is in this bin.
      // If so, display it in the approriate color
      int c = 0;              
      for (int i = 0; i < numSamplesToDisplay; i++) {  // loop through the required number of samples to display
        int sID = samples.get( (dIdx+i) % (samples.size()) );
        int sIdx = m_sampleIDs.indexOf(sID);
        if (sIdx >= 0) {                               // this sample appears in this bin, so display it!
          numInBin++;
          int y = m_y - (numInBin*m_w) + m_w/2;
          if (m_node.m_model.m_ssDisplayMode == 1 || m_node.m_model.m_ssDisplayMode == 2)
            c = color(sortedHues.get(numInBin-1), 255, 255);
          else
            c = color(m_node.m_model.getSampleHue(dIdx+i), 255, 255);
          fill(c);
          ellipse(m_x + (m_w/2), y, m_w, m_w);
        }
      }
    }
    
    popStyle();    
  }
  
  
  public void drawColourBins() {
    // helper method for drawing
    // draws sections of bin in different colours in ColourBins mode
    
    pushStyle();
    
    fill(m_focalOutsideRangeSelFillColor);
    rect(m_x, m_y-m_h, m_w, m_h); 
     
    colorMode(HSB);
    stroke(0);

    ArrayList<Integer> samples = m_node.m_model.m_ssSamplesToDisplay;     // a list of all sampleIDs currently selected
                
    if (!samples.isEmpty()) {
      
      // We need to first work out the color of each sample to be displayed in this bin, 
      // then sort the colors so that we display contiguous blocks of each color.
      ArrayList<Integer> sortedHues = new ArrayList<Integer>();
      int idx = 0;
      for (Integer id : samples) { 
        int sIdx = m_sampleIDs.indexOf(id.intValue());
        if (sIdx >= 0) {                             // this sample appears in this bin, so record its color               
          sortedHues.add(m_node.m_model.getSampleHue(idx));
        }
        idx++;
      }
      Collections.sort(sortedHues);
      
      if (!sortedHues.isEmpty()) {
        int numSelectedSamples = sortedHues.size();
        ArrayList<Integer> hues = new ArrayList<Integer>();
        ArrayList<Integer> counts = new ArrayList<Integer>();
        int currentHue = sortedHues.get(0);
        int currentCount = 0;
        for (Integer hue : sortedHues) {
          if (hue != currentHue) {
            hues.add(currentHue);
            counts.add(currentCount);
            currentHue = hue;
            currentCount = 1;
          }
          else {
            currentCount++;
          }
        }
        hues.add(currentHue);
        counts.add(currentCount);
        
        //println((m_node.hasFocus())?"FOCUS NODE*********************":"NON-FOCUS NODE");
        /*
        if (m_node.m_name.equals("cp")) {
          println(m_node.m_name);
          println(numSelectedSamples + " :: " + hues.size() + " :: " + sortedHues.size());
        }
        */
        
        int dy = 0;
        int dh = 0;
        int fullh = (m_h * numSelectedSamples) / m_numSamples;
        int c = 0;
        for (int i=0; i < hues.size(); i++) {
          c = color(hues.get(i), 255, 255);
          dh = round(((float)(counts.get(i) * fullh)) / (float)numSelectedSamples);
          /*
          if (m_node.m_name.equals("cp")) {
            println(i+": "+counts.get(i)+", "+numSelectedSamples+", "+dh+", "+dy);
          }
          */
          fill(c);
          rect(m_x, m_y-dy-dh, m_w, dh);
          dy += dh;
        }
      }
    }
    
    popStyle();    
  }  
  
  
  public void drawBrushedBin(int numBrushes) {
    // helper method for drawing
    // draws the various sections of an individual bin according to the current brushing of the bin
    
    int cumH = 0; // cumulative count of height of brushes considered so far 
    int cumB = 0; // cumulative count of total number of brushed samples
    
    for (int b=0; b<numBrushes; b++) {
      float brushFrac = (float)m_numBrushedSamples[b] / (float)m_numSamples;
      int brushH = round(brushFrac*(float)m_h);
      cumB += m_numBrushedSamples[b];
      
      // add brush height to cumulative total, but check we don't overshoot or undershoot the
      // max total height of the bin because of rounding errors 
      if (cumH + brushH > m_h) {
        brushH = m_h - cumH;
        cumH = m_h;
      }
      else if (cumB == m_numSamples) {
        brushH = m_h - cumH;
        cumH = m_h;
      }
      else {
        cumH += brushH;
      }

      // draw brushed section
      if (brushH > 0) {
        fill(m_brushColors[b]);
        rect(m_x, m_y-cumH, m_w, brushH);
      }
    }

    // draw unbrushed section
    fill(m_nonFocalNonBrushedFillColor);
    rect(m_x, m_y-m_h, m_w, m_h-cumH);
  }

  
  public boolean inSelectedRange() {
    // is this bin within the range currently selected by the Range Selector slider?
    return ((m_idx >= m_node.m_rsLow) && (m_idx <= m_node.m_rsHigh));
  }
  
  
  public int numSamples() {
    return m_numSamples;
    //return m_sampleIDs.size();
  }  
  
  
  public int numBrushed() {
    // return the number of brushed samples (with no misses)
    assert(m_numBrushedSamples[0] == m_brushedSampleIDs.size());
    return m_numBrushedSamples[0];
  }
  
  
  public void brushSamples(ArrayList<Integer> samples) {
    // For single-node brushing
    // Calculate which of the samples passed in are members of this bin, and record that number
    // in m_numBrushedSamples[0]
    
    ArrayList<Integer> smallList;
    ArrayList<Integer> bigList;
    
    if (samples.size() > m_sampleIDs.size()) {
      smallList = m_sampleIDs;
      bigList = samples;
    }
    else {
      smallList = samples;
      bigList = m_sampleIDs;
    }
    
    int matches = 0;
    m_brushedSampleIDs.clear();
    
    for (Integer sampleID : smallList) {
      if (bigList.contains(sampleID)) {
        matches++;
        assert(!m_brushedSampleIDs.contains(sampleID));
        m_brushedSampleIDs.add(sampleID);
      }
    }
    
    m_numBrushedSamples[0] = matches;
  }
  

  public boolean brushSampleAdd(int sampleID, int numMisses) {
    // If the sample passed in is in this bin, increment the relevant count of brushed tiles
    // and return true, otherwise return false
    
    assert((numMisses >= 0) && (numMisses < m_node.m_hgNumBrushes));
    
    if (m_sampleIDs.contains(sampleID)) {
      m_numBrushedSamples[numMisses]++;
      if (numMisses == 0) {
        m_brushedSampleIDs.add(sampleID);
      }
      return true;
    }
    else {
      return false;
    }
  }
  
  
  public void resetBrushing() { 
    for (int i=0; i<m_node.m_hgNumBrushes; i++) {
      m_numBrushedSamples[i] = 0;
    }
    m_brushedSampleIDs.clear();
  }
  

  public boolean sampleBrushed(int sampleID) {
    // returns true if the specified sample is brushed (numMisses==0) in this bin   
    return m_brushedSampleIDs.contains(sampleID);    
  }  
  
  
  public boolean sampleInBin(int sampleID) {
    return m_sampleIDs.contains(sampleID);
  }
  
  
  public float getTotalValues() {
    // Return the sum of values of samples in this bin.
    return getTotalValues(false);
  }
  
  
  public float getTotalValues(boolean brushedOnly) {
    // Return the sum of values of samples in this bin.
    // If brushedOnly==true, only count brushed samples (with numMiss==0), otherwise count all samples.
  
    ArrayList<ArrayList<Number>> dataArray = m_node.m_model.m_data;
    int col = m_node.m_dataArrayCol;
    float total = 0.0f;
    
    if (brushedOnly) {
      for (Integer id : m_sampleIDs) {
        if (sampleBrushed(id)) {
          total += dataArray.get(id).get(col).floatValue();
        }
      }
    }
    else {
      for (Integer id : m_sampleIDs) {
        total += dataArray.get(id).get(col).floatValue();
      }
    }
    
    return total;
  }
  
   
  public ArrayList<Float> getSelectedValues() {
    // Return a list of values of all samples in this bin.
    return getSelectedValues(false);
  }
  
  
  public ArrayList<Float> getSelectedValues(boolean brushedOnly) {
    // Return a list of values of samples in this bin.
    // If brushedOnly==true, only include brushed samples (with numMiss==0), otherwise include all samples.
    
    ArrayList<Float> values = new ArrayList<Float>();
    
    ArrayList<ArrayList<Number>> dataArray = m_node.m_model.m_data;
    int col = m_node.m_dataArrayCol;
    
    if (brushedOnly) {
      for (Integer id : m_sampleIDs) {
        if (sampleBrushed(id)) {
        values.add(dataArray.get(id).get(col).floatValue());
        }
      }
    }
    else {
      for (Integer id : m_sampleIDs) {
        values.add(dataArray.get(id).get(col).floatValue());
      }
    }
    
    return values;    

  }
  
}





public class Model {

  ArrayList<ArrayList<Number>> m_data; // holds the model's data, indexed by m_data.get(sampleID).get(node.m_dataArrayCol)
  ArrayList<String> m_dataLabels;
  
  // general data about the model
  String  m_modelName;
  boolean m_modelHasData;
  boolean m_modelHasDataLabels;
  int     m_modelDataLabelFileCol;  // column in data file corresponding to sample label (1-based)
  String  m_modelDataFilename;
  boolean m_modelLiveRun;
  int     m_rngSeed;
  
  // information about individual nodes
  ArrayList<Node> m_rnodes; // Root nodes
  ArrayList<Node> m_inodes; // Intermediate nodes
  ArrayList<Node> m_lnodes; // Leaf nodes

  ArrayList<Node> m_allNodes;     // contains ALL nodes (the union of rnodes, inodes and lnodes)
  boolean m_allNodesSafe = false; // can we use allNodes or do we have to rebuild it?
  
  // a list of all currently-selected samples
  HashSet<Integer> m_allSelectedSamples; // sample IDs (index of m_data) of all samples currently selected in focal nodes
  
  // information about brush links between nodes
  ArrayList<BrushLink> m_brushLinks;
  boolean m_bNewBrushLinkUnderConstruction;
  Node m_newBrushLinkNode1;
  
  // information about causal links between nodes
  ArrayList<CausalLink> m_causalLinks;
  
  // information about minimised nodes
  ArrayList<Node> m_minimisedNodes;
  
  // current mode of interaction
  InteractionMode m_interactionMode = InteractionMode.Unassigned;
  
  // information needed in ShowSamples interaction mode
  boolean m_ssAutoUpdate = true;           // flags whether samples are automatically or manually updated
  int m_ssTimer = 0;                       // a timer to determine when a new sample should be shown
  int m_ssTimerReset = 30;                 // timer value at which a new sample is shown and timer reset occurs
  int m_ssMaxSamplesToDisplay = 5;         // maximum allowed length of m_ssSamplesToDisplay 
  ArrayList<Integer> m_ssSamplesToDisplay; // a list of all samples currently in m_allSelectedSamples, in random order
  int m_ssDisplayIdx = 0;                  // index of first currently displayed sample in m_ssSamplesToDisplay
  ArrayList<Integer> m_ssSampleHues;       // list of color hues used for displaying samples
  int m_ssDisplayMode = 2;                 // 0=assign colors to samples at random, 
                                           // 1=assign according to bin in focal node (colours random for adjacent bins),
                                           // 2=assign according to bin in focal node (colours progress for adjacent bins)
  
  // visualisation options
  boolean m_visTiled = true;               // display histogram bins as tiled or continuous?
  boolean m_showMedians = false;           // display histogram medians rather than means in range selector bar?
  boolean m_showCausalLinks = true;        // display causal links?
  
  // global information about appearance
  int   m_globalZoom = 100;
  int   m_nodeDefaultHeight  = 200;
  int   m_nodeDefaultWidth   = 330;
  float m_nodeBinScaleFactor = 5.0f; // used to determine the default scale of histogram bins in nodes
  int   m_minInternodeGap = 20;
  int   m_minimisedInternodeGap = 5;
  int   m_numRootCols = 1;
  int   m_numInterCols = 1;
  int   m_numLeafCols = 1;
  
  // global UI items
  int     m_menuH = 50;
  boolean m_bShowHelpScreen = false;
  boolean m_bShowModeMessage = false;
  int     m_modeMessageTimer = 0;
  int     m_modeMessageDuration = 50; // length of time (in number of frames) to show a mode message
  int   m_menuBackgroundColor = 0xFF000000;
  int   m_menuTextColor = 0xFFEEEE30;
  int   m_windowBackgroundColor = 0xFF808080;
  PFont   m_smallFont;
  PFont   m_mediumFont;
  int     m_defaultInterNodeGapV = 20;
  String  m_strMu = "\u03BC";
  String  m_strXTilde = "x\u0303";

  
  //////////// METHODS //////////////////
  
  Model(String configXMLfilename) {
    
    setDefaults();
     
    XML xml = null;
   
    try {
      xml = loadXML(configXMLfilename);
    } 
    catch (Exception e) {
      println("Input file does not appear to be an XML file of the correct format. Aborting!");
      exit();
    }
    
    if (xml == null) {
      println("Problem loading XML file. Aborting!");
      exit();
    }   
    
    XML general = xml.getChild("general");
    if (general == null) {
      println("Could not find general section in XML file. Aborting!");
      exit();
    }
    
    m_modelDataFilename = general.getString("data");
    m_modelHasData = (m_modelDataFilename != null);

    String sDataHasLabels = general.getString("has-labels");
    m_modelHasDataLabels = (sDataHasLabels != null) && (sDataHasLabels.equals("true"));
    
    m_modelDataLabelFileCol = general.getInt("label-filecol");
    
    String sLiveRun = general.getString("live");
    m_modelLiveRun = (sLiveRun != null) && (sLiveRun.equals("true"));
    
    m_rngSeed = general.getInt("rng-seed", millis());
    if (m_rngSeed < 0) {
      m_rngSeed = millis();
    }
    randomSeed(m_rngSeed);
    
    int numSamplesFromFile = general.getInt("num-samples", -1);
    
    XML modelLabel = general.getChild("label");
    if (modelLabel != null) {
      m_modelName = modelLabel.getContent();
    }
    
    XML appearance = xml.getChild("appearance");
    if (appearance != null) {
      m_nodeDefaultHeight = appearance.getInt("node-default-height", m_nodeDefaultHeight);
      m_nodeDefaultWidth = appearance.getInt("node-default-width", m_nodeDefaultWidth);
      m_nodeBinScaleFactor = appearance.getFloat("node-bin-scale-factor", m_nodeBinScaleFactor);
      m_minInternodeGap = appearance.getInt("min-internode-gap", m_minInternodeGap);
      m_numRootCols = appearance.getInt("num-root-cols", m_numRootCols);
      m_numInterCols = appearance.getInt("num-inter-cols", m_numInterCols);
      m_numLeafCols = appearance.getInt("num-leaf-cols", m_numLeafCols);
      String showCausalLinks = appearance.getString("show-causal-links");
      m_showCausalLinks = (showCausalLinks != null) && showCausalLinks.equals("true");
    }
    
    
    // Now go through each of the nodes specified in the XML file, and create a new Node object of
    // the appropriate subclass (discrete, continuous) for each one
    
    XML nodelist = xml.getChild("nodes");
    if (nodelist == null) {
      println("No data nodes are defined in the XML file. Aborting!");
      exit();
    }
    
    XML[] nodes = nodelist.getChildren("node");
    
    int rYpos = 20;
    int iYpos = 20;
    int lYpos = 20;   
   
    int nextFreeDataCol = 0; 
    
    for (int i=0; i < nodes.length; i++) {
      // process each node in turn
      
      int imin, imax;
      float fmin, fmax;
      ArrayList<Integer> parentIDs = new ArrayList<Integer>();
      Node newnode = null;
      
      XML xnode = nodes[i];
      
      int id = xnode.getInt("id", 0);
      int filecol = xnode.getInt("filecol", 0);
      
      String strRole = xnode.getString("role"); // this determines whether node is placed into rnodes, inodes or lnodes
      if (strRole == null) {
        println("No role attribute specified for node "+id+" in XML file!");
        exit();
      }
      int role = 0;
      if (strRole.equals("root"))
        role = 0;
      else if (strRole.equals("inter"))
        role = 1;
      else if (strRole.equals("leaf"))
        role = 2;
      else {
        println("Oops! Found a node with unknown role '" + strRole + "' in file " + configXMLfilename);
        exit();
      }      

      String name = "";
      XML label = xnode.getChild("label");
      if (label != null) {
        name = label.getContent();
      }

      XML parentlist = xnode.getChild("parents");
      if (parentlist != null) {
        XML[] parents = parentlist.getChildren("parent");
        for (int j=0; j < parents.length; j++) {
          int pid = parents[j].getInt("id");
          parentIDs.add(pid);
        }
      }
      
      String datatype = xnode.getString("datatype");
      if (datatype.equals("discrete")) {
        imin = xnode.getInt("min");
        imax = xnode.getInt("max");
        newnode = new DiscreteNode(this,id,name,filecol,nextFreeDataCol++,imin,imax,role,parentIDs);
      }
      else if (datatype.equals("continuous")) {
        fmin = xnode.getFloat("min");
        fmax = xnode.getFloat("max");
        newnode = new ContinuousNode(this,id,name,filecol,nextFreeDataCol++,fmin,fmax,role,parentIDs);
      }
      else {
        println("Oops! Found a node of unknown type '" + datatype + "' in file " + configXMLfilename);
        exit();
      }
      
      switch (role) {
        case 0: {
          m_rnodes.add(newnode);
          //println("Adding new rnode " + newnode.m_name);
          break;
        }
        case 1: {
          m_inodes.add(newnode);
          //println("Adding new inode " + newnode.m_name);
          break;
        }
        case 2: {
          m_lnodes.add(newnode);
          //println("Adding new lnode " + newnode.m_name);
          break;
        }
        default: {
          println("Oops! Found a node with unknown role '" + role + "' in file " + configXMLfilename);
          exit();
        }
      }
    }
    
    // Having processed all of the nodes, set up any causal links as specified by each node's
    // list of parent nodes
    initialiseCausalLinks();
    
    // We have now created all of the nodes, now we lay them out on the screen according to any
    // layout specifications given in the XML file (note that later on we tweek the positions
    // once we have loaded in the data, in case any node has had to be resized)
    tileNodes(); 
    
    // Having read all of the model specification file, load in the data if a
    // data file has been specified
    if (m_modelHasData) {
      
      // load data
      load(m_modelDataFilename, numSamplesFromFile);
      
      // first pass of initialisation of nodes based upon the data associated with each one
      checkAllNodesSafe();      
      int numNodes = m_allNodes.size(); 
      int[] nodeHeights = new int[numNodes]; 
      int n = 0;
      for (Node node : m_allNodes) {
        node.initialiseHistogram();
        nodeHeights[n++] = node.getH();
      }
      
      // having initialised nodes, rescale them according to the distribution of heights of all nodes
      Arrays.sort(nodeHeights);
      int defaultH = m_allNodes.get(0).getDefaultH();
      int refH = nodeHeights[(numNodes*3)/4];
      if (defaultH != refH) {
        float sf = (float)refH / (float)defaultH;
        for (Node node : m_allNodes) {
          node.setH((int)(((float)node.getH())*sf), true, true);
        }
      }
      
      // The node heights might have been adjusted now that we have added the data, so retile all 
      // nodes to ensure that they are still evenly spaced
      tileNodes();
    }
  }
  
  
  public void setDefaults() {
    m_data               = new ArrayList<ArrayList<Number>>();
    m_dataLabels         = new ArrayList<String>();
    m_rnodes             = new ArrayList<Node>();
    m_inodes             = new ArrayList<Node>();
    m_lnodes             = new ArrayList<Node>();
    m_allNodes           = new ArrayList<Node>();
    m_allNodesSafe       = false;
    m_brushLinks         = new ArrayList<BrushLink>();
    m_bNewBrushLinkUnderConstruction = false;
    m_causalLinks        = new ArrayList<CausalLink>();
    m_minimisedNodes     = new ArrayList<Node>();
    m_allSelectedSamples = new HashSet<Integer>();
    m_ssSamplesToDisplay = new ArrayList<Integer>();
    m_ssSampleHues       = new ArrayList<Integer>();
    m_modelDataLabelFileCol = 0;
    m_smallFont          = createFont("Arial", 11, true);
    m_mediumFont         = createFont("Arial", 16, true);
  }


  public void load(String filename, int numSamplesToPick) {
    // Having set up the Node objects, read in data from the specified data file (in CSV format)
    // and store it in the m_data variable.
    // If numSamplesToPick > 0, we pick that number of samples at random from the file, otherwise
    // we load all samples.
    
    String lines[] = loadStrings(filename);
    
    checkAllNodesSafe();
    int numNodes = m_allNodes.size();
    
    int numLinesInFile = lines.length;
    
    ArrayList<Integer> linesToLoad = new ArrayList<Integer>();
    
    // If we are picking specific lines, create a list of lines to pick, selected at random
    if ((numSamplesToPick > 0) && (numSamplesToPick < numLinesInFile)) {
      ArrayList<Integer> linesAvailable = new ArrayList<Integer>(numLinesInFile);
      for (int i = 0; i < numLinesInFile; i++) {
        linesAvailable.add(i);
      }
      for (int i = 0; i < numSamplesToPick; i++) {
        linesToLoad.add( linesAvailable.remove( (int)random(linesAvailable.size()) ) );
      }
    }
    
    // cycle through each line of data in the data file
    for (int i = 0 ; i < numLinesInFile; i++) {
      
      // if we are picking specific lines and this isn't one of them, skip to the next line!
      if ((!linesToLoad.isEmpty()) && (!linesToLoad.contains(i))) {
        continue; 
      }

      boolean dataAdded = false;
      boolean exception = false;
      String label = "";
      
      String[] data = splitTokens(lines[i],",");
      ArrayList<Number> row = new ArrayList<Number>(numNodes);
      for (int n=0; n<numNodes; n++) {
        // just fill the array with 0s for now. We'll replace these with the real data in due course
        row.add(0); 
      }
        
      // read each column of data in this row
      for (int j=0; j < data.length; j++) {
      
        int thisFileCol = j+1;
        
        // work out where to store this data
        for (Node node : m_allNodes) {

          if (node.m_dataFileCol == thisFileCol) {
            
            if (m_modelHasDataLabels && (m_modelDataLabelFileCol == thisFileCol)) {
              label = data[j];
            }
            else if (node instanceof DiscreteNode) {
              try {
                Integer val = Integer.valueOf(data[j]);
                row.set(node.m_dataArrayCol, val);
                dataAdded = true;
              }
              catch (Exception e) {
                exception = true;
              }
            }  
            else if (node instanceof ContinuousNode) {
              try {
                Float val = Float.valueOf(data[j]);
                row.set(node.m_dataArrayCol, val);
                dataAdded = true;
              }
              catch (Exception e) {
                exception = true;
              }
            }
            else {
              // Don't know what to do with this column of data, so ignore it
            }
          }
        }
      }
      
      // if we successfully read all of the data in the row, then store it!
      if (exception) {
        println("Problem encountered in line "+(i+1)+" of data file. Ignoring data in this line.");
      }
      else if (dataAdded) {  
        m_data.add(row);
        m_dataLabels.add(label);
      }
      
    }

  }

  
  public void checkAllNodesSafe() {
    // m_allNodes contains an aggregrate list of all root, intermediate and leaf nodes in a single list
    // If any of these lists are changed (for example a node is move from one list to another, or a node
    // is added or deleted), then the program should set the m_allNodesSafe flag to false. Then, whenever
    // the program wants to use m_allNodesSafe, it should first call this method. This checks whether it is
    // safe to use the list, and if not, rebuilds the list to reflect the current state of all nodes.
    //
    // In practice, none of these scenarios of changing the contents of root, intermediate and leaf node lists
    // is currently implemented, so this method is not actually required at this stage!
    
    if (!m_allNodesSafe) {
      m_allNodes.clear();
      m_allNodes.addAll(m_rnodes);
      m_allNodes.addAll(m_inodes);
      m_allNodes.addAll(m_lnodes);
      m_allNodesSafe = true;
    }
  }
  
  
  public void initialiseCausalLinks() {
    // Look at the parents of each node, and create a new CausalLink object for each link so specified and
    // add it to m_causalLinks
    
    m_causalLinks.clear();
    
    checkAllNodesSafe();
    for (Node nodeTo : m_allNodes) {
      for (Integer nodeFromID : nodeTo.m_parentIDs) {
        Node nodeFrom = getNodeFromID(nodeFromID);
        if (nodeFrom != null) {
          m_causalLinks.add( new CausalLink(this, nodeFrom, nodeTo) );
        }
      }
    }
  }
  
  
  public Node getNodeFromID(int id) {
    // Find the node specified by the given id (Node.m_id) and return it, otherwise return null if
    // no matching node is found.
    
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      if (node.m_id == id) {
        return node;
      }
    }
    
    return null;
  }
  
  
  public void tileNodes() {
    // Arrange all nodes to have regular spacing, according to the node heights and the
    // model variables m_numRootCols, m_numInterCols, m_numLeafCols, m_minInternodeGap
    
    int currentColXpos = m_minInternodeGap;
    
    if (!m_rnodes.isEmpty()) {
      currentColXpos = layoutNodeColumns(m_rnodes, m_numRootCols, currentColXpos);   
      currentColXpos += m_minInternodeGap;
    }
    
    if (!m_inodes.isEmpty()) {
      currentColXpos = layoutNodeColumns(m_inodes, m_numInterCols, currentColXpos);   
      currentColXpos += m_minInternodeGap;
    }
    
    if (!m_lnodes.isEmpty()) {
      currentColXpos = layoutNodeColumns(m_lnodes, m_numLeafCols, currentColXpos);   
    }
  }
  
  
  public int layoutNodeColumns(ArrayList<Node> nodes, int numCols, int firstColXpos) {
    // Helper method for tileNodes
    // Lay out the nodes passed in into the specified number of columns, beginning the
    // first column at x position firstColXpos. 
    // The method returns the x position to the right of the final column laid out
    // plus a small extra spacing gap.
  
    int currentColXpos = firstColXpos;
    int currentColYpos = m_minInternodeGap;  
    int nodesPerCol = ceil((float)nodes.size() / (float)numCols);
    int nIdx = 0;
    
    for (int c = 0; c < numCols; c++) {
      int maxW = 0; // track the maximum width of a node in the current column
      for (int n = 0; n < nodesPerCol && nIdx < nodes.size(); n++, nIdx++) {
        Node node = nodes.get(nIdx);
        node.setPosition(currentColXpos, currentColYpos);
        currentColYpos += (node.getH() + m_minInternodeGap);
        maxW = max(maxW, node.getW());
      }
      currentColXpos += (maxW + m_minInternodeGap);
      currentColYpos = m_minInternodeGap;
    }
    
    return currentColXpos;
  }
  
  
  public void setSingleFocus(int nodeIdx) {
    // in SingleNodeBrushing mode, set the indicated node to be the focus node, and
    // remove focus from all other nodes
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      if (node.m_id == nodeIdx) {
        node.m_bHasFocus = true;
      }
      else {
        node.m_bHasFocus = false;
        node.setFullRange();
      }
    }
  }
  
  
  public void toggleMultiFocus(int nodeIdx) {
    // in MultiNodeBrushing mode, toggle the focus flag of the indicated node
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      if (node.m_id == nodeIdx) {
        node.m_bHasFocus = !node.m_bHasFocus;
        if (!node.m_bHasFocus) {
          node.setFullRange();
        }
        break;
      }
    }
  }  
  
  
  public void setInteractionMode(InteractionMode mode) {
    switch (mode) {
      case SingleNodeBrushing: {
        m_interactionMode = InteractionMode.SingleNodeBrushing;
        resetAllBrushing();
        checkAllNodesSafe();
        // now ensure that at most one Node has the focus
        for (Node node : m_allNodes) {
          if (node.m_bHasFocus) {
            setSingleFocus(node.m_id);
            brushAllNodesOnOneSelection(node);
            redraw();
            break;
          }
        }
        break;  
      }
      case MultiNodeBrushing: {
        m_interactionMode = InteractionMode.MultiNodeBrushing;
        resetAllBrushing();
        brushAllNodesOnMultiSelection();
        redraw();
        break;
      }
      case ColourBins: {
        m_interactionMode = InteractionMode.ColourBins;
        resetAllBrushing();
        checkAllNodesSafe();
        // now ensure that at most one Node has the focus
        for (Node node : m_allNodes) {
          if (node.m_bHasFocus) {
            setSingleFocus(node.m_id);
            brushAllNodesOnOneSelection(node);
            //redraw();
            break;
          }
        }        
        //m_ssTimer = 0;
        updateSelectedSampleList();
        redraw();
        break;        
      }
      case ShowSamples: {
        m_interactionMode = InteractionMode.ShowSamples;
        resetAllBrushing();
        //brushAllNodesOnMultiSelection();
        checkAllNodesSafe();
        // now ensure that at most one Node has the focus
        for (Node node : m_allNodes) {
          if (node.m_bHasFocus) {
            setSingleFocus(node.m_id);
            brushAllNodesOnOneSelection(node);
            //redraw();
            break;
          }
        }        
        m_ssTimer = 0;
        updateSelectedSampleList();
        redraw();
        break;
      }
      case Unassigned:
      default: {
        println("Unexpected Interaction Mode in Model.setInteractionMode!");
      }
    }
    
    m_bShowModeMessage = true;
    m_modeMessageTimer = m_modeMessageDuration;

  }
  
  
  public void brushAllNodesOnOneSelection(Node focalNode) {
    // first gather a list of selected samples
    ArrayList<Integer> selectedSampleIDs = focalNode.getSelectedSampleIDs();
    for (Node node : m_allNodes) {
      if (node != focalNode) {
        node.brushSamples(selectedSampleIDs);
      }
    }
  }
  
  
  public void brushAllNodesOnMultiSelection() {
    
    // First build up a list of all nodes that currently have focus, and a list of all other nodes
    
    ArrayList<Node> focalNodes = new ArrayList<Node>();
    ArrayList<Node> otherNodes = new ArrayList<Node>();
    
    for (Node node : m_allNodes) {
      if (!node.minimised()) {
        if (node.m_bHasFocus) {
          focalNodes.add(node);
        }
        else {
          otherNodes.add(node);
        }
      }
    }
    
    for (Node onode : otherNodes) {
      onode.resetBrushing();
    }

    // Now go through every sample in the data and check whether it lies within the selected range of values
    // for each of the focal nodes (and if not, how many focal nodes it misses). Then brush the sample
    // in all the other nodes according to the number of misses.

    int numBrushes = m_allNodes.get(0).m_hgNumBrushes;
    int numSamplesAll = m_data.size();
    
    for (int i=0; i<numSamplesAll; i++) {
      int numMisses = 0;
      
      for (Node fnode : focalNodes) {
        if (!(fnode.sampleSelected(i))) {
          numMisses++;
        }
      }
      
      if (numMisses < numBrushes) {
        for (Node onode : otherNodes) {
          onode.brushSampleAdd(i, numMisses);
        }
      }
    }
    
  }  
  
  
  public void resetAllBrushing() {
    for (Node node : m_allNodes) {
      node.resetBrushing();
    }    
  }
  
  
  public void draw(int globalZoom, int nodeZoom) {
    
    background(m_windowBackgroundColor);
    
    m_globalZoom = globalZoom;
    
    // draw brush links
    for (BrushLink bLink : m_brushLinks) {
      bLink.draw();
    }
    
    // draw causal links if required
    if (m_showCausalLinks) {
      for (CausalLink cLink : m_causalLinks) {
        cLink.draw();
      }
    }
    
    // draw nodes
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      node.draw(nodeZoom);
    }
    
    // draw minimised nodes
    drawMinimisedNodes();
       
    // update ShowSamples data if required
    if (m_interactionMode == InteractionMode.ShowSamples) {
      updateShowSamples();
    }
    
    // update the Mode Message timer
    if (m_modeMessageTimer > 0) {
      m_modeMessageTimer--;
      if (m_modeMessageTimer == 0) {
        m_bShowModeMessage = false;
      }
    }
      
    if (m_bShowModeMessage) {
      int mw = 200;
      int mh = 75;
      int mx = (width-mw)/2;
      int my = (height-mh)/2;
      fill(m_menuBackgroundColor);
      rect(mx, my, mw, mh);
      String mode = getInteractionModeStr();
      textFont(m_mediumFont, 16);
      textAlign(CENTER);
      fill(m_menuTextColor);
      text(mode, mx+(mw/2), my+(mh/2)+6);
    }
    
    // draw help screen if required
    if (m_bShowHelpScreen) {
      drawHelpScreen();
    }
    
  }
  
  
  public String getInteractionModeStr() {
    String mode;
    switch (m_interactionMode) {
      case SingleNodeBrushing:
        mode = "Single Node Brushing";
        break;
      case MultiNodeBrushing:
        mode = "Multi Node Brushing";
        break;
      case ColourBins:
        mode = "Omnibrushing";
        break;
      case ShowSamples:
        mode = "Sample View";
        break;
      default:
        mode = "(no mode set)";
    }
    return mode;
  }
  
  
  public void drawHelpScreen() {
    pushStyle();
    pushMatrix();
    
    // fade out the current screen content
    fill(100, 100);
    rect(0, 0, width, height);
    
    // draw a background for the help screen
    fill(0);
    rect(50, 10, width-100, height-20);
    translate(50, 25);
    
    textFont(m_mediumFont, 16);
    textAlign(LEFT);
    fill(m_menuTextColor);
    
    int x1 = 10;
    int x2 = 310;
    int x2a = 460;
    int x3 = 610;
    int y = 15;
    int dy = 25;
    int bigdy = 50;
    
    int headingTextColor = 0xff9090D0;
    
    String modeStr = getInteractionModeStr();
    
    fill(headingTextColor);
    text("INTERACTION MODES", x1, y);
    y += dy;
    fill(m_menuTextColor);
    text("'1' = Single Node Brushing", x1, y);
    text("'2' = Multi Node Brushing", x2a, y);
    y += dy;
    text("'3' = Omnibrushing", x1, y);
    text("'4' = Sample View", x2a, y);
    y += dy;
    text("[Current mode: "+modeStr+"]", x1, y);
    y += bigdy;
    fill(headingTextColor);
    text("ZOOM DISPLAY", x1, y);
    y += dy;
    fill(m_menuTextColor);
    text("'+' = Zoom In", x1, y);
    text("'-' = Zoom Out", x2a, y);
    y += bigdy;
    fill(headingTextColor);
    text("GENERAL ACTIONS", x1, y);
    y += dy;
    fill(m_menuTextColor);
    text("'H' = Toggle Help Screen On/Off", x1, y);
    text("'M' = Toggle Display of Mean/Median for Nodes", x2a, y);
    y += bigdy;
    fill(headingTextColor);
    text("ACTIONS IN 'SHOW INDIVIDUAL SAMPLES' MODE", x1, y);
    y += dy;
    fill(m_menuTextColor);
    text("'ARROW UP' = Speed Up Automatic Sample Cycling", x1, y);
    text("'ARROW DOWN' = Slow Down Automatic Sample Cycling", x2a, y);
    y += dy;
    text("'ARROW LEFT' = Single Step Backward Through Samples", x1, y);
    text("'ARROW RIGHT' = Single Step Forward Through Samples", x2a, y);
    y += dy;
    text("'Z' = Decrease Number of Samples", x1, y);
    text("'X' = Increase Number of Samples", x2a, y);
    y += dy;
    text("'S' = Switch between sample colouring modes: random / by bin (bin colours random) / by bin (bin colours ordered)", x1, y);
    y += bigdy;
    fill(headingTextColor);
    text("BRUSH LINKS", x1, y);
    y += dy;
    fill(m_menuTextColor);
    text("'L' = Create New Brush Link (on Focal Nodes only)", x1, y);
    text("'B' = Delete Brush Link (over link handle)", x2a, y);
    y += bigdy;
    fill(headingTextColor);
    text("CAUSAL LINKS", x1, y);
    y += dy;
    fill(m_menuTextColor);
    text("'D' = Toggle Causal Link Direction (over link handle)", x1, y);
    text("'C' = Delete Causal Link (over link handle)", x2a, y);
    y += dy;
    text("'R' = Reinitialise Causal Links", x1, y);
    
    popMatrix();
    popStyle();
  }
  
  
  public void updateShowSamples() {
    // In ShowSample interaction mode, we need to update a timer at each frame, and, at regular intervals,
    // chose a new sample from the currently selected samples to add to the display list. We also need
    // to remove old samples from the end of the list.
    
    if (m_ssAutoUpdate) {
      m_ssTimer++;
      if (m_ssTimer >= m_ssTimerReset) {
        m_ssTimer = 0;  
        if (!m_ssSamplesToDisplay.isEmpty()) {
          m_ssDisplayIdx = (m_ssDisplayIdx+1) % m_ssSamplesToDisplay.size();
        }  
      }
    }
  }

  
  public void updateSelectedSampleList() {
    // Update m_allSelectedSamples to contain a list of all Samples currently selected by the range selectors
    // of all focal nodes.
    // Having done that, update m_ssSamplesToDisplay to be a random permutation of the new m_allSelectedSamples.
    
    m_allSelectedSamples.clear();
    boolean firstFocalNode = true;
    Node ffNode = null;
    
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      if (node.hasFocus()) {
        if (firstFocalNode) {
          m_allSelectedSamples.addAll(node.getSamplesInRange());
          ffNode = node;
          firstFocalNode = false;
        }
        else {
          m_allSelectedSamples.retainAll(node.getSamplesInRange()); // keep the intersection of this and previous sets
        }
      }
    }

    // Reset m_ssSamplesToDisplay to be an ArrayList corresponding to the new m_allSelectedSamples HashSet
    m_ssSamplesToDisplay = new ArrayList<Integer>(m_allSelectedSamples);
    Collections.shuffle(m_ssSamplesToDisplay);
    m_ssDisplayIdx = 0;
    
    // And finally create color palette for the samples
    boolean matchColoursToBins = false;
    if (m_interactionMode == InteractionMode.ShowSamples && (m_ssDisplayMode == 1 || m_ssDisplayMode == 2))
      matchColoursToBins = true;
    else if (m_interactionMode == InteractionMode.ColourBins)
      matchColoursToBins = true;
      
    if (matchColoursToBins && ffNode != null) {
      // assign hue for each sample according to which bin it belongs to in the focal node
      ffNode.matchSampleBinsToColors(m_ssSamplesToDisplay, m_ssSampleHues);
    }
    else {
      // assign hue for each sample at random
      int numSamples = m_ssSamplesToDisplay.size();
      m_ssSampleHues.clear();
      for (int i = 0; i < numSamples; i++) {
        m_ssSampleHues.add((i*255)/numSamples);
      }
      Collections.shuffle(m_ssSampleHues);      
    }
    
  }
  
  
  public void showSamplesAutoSpeedUp() {
    // Decrease the time between displaying new samples in ShowSamples mode when under automatic control
    m_ssAutoUpdate = true;
    if (m_ssTimerReset > 1) {
      m_ssTimerReset--;
    }
  }
  
  
  public void showSamplesAutoSlowDown() {
    // Increase the time between displaying new samples in ShowSamples mode when under automatic control
    m_ssAutoUpdate = true;
    if (m_ssTimerReset < 100) {
      m_ssTimerReset++;
    }
  }
  
  
  public void showSamplesStepBackward() {
    // Reverse to previous sample in ShowSamples mode when under manual user control
    m_ssAutoUpdate = false;
    if (m_ssSamplesToDisplay.isEmpty()) {
      m_ssDisplayIdx = 0;
    }
    else {
      m_ssDisplayIdx = (m_ssDisplayIdx <= 0) ? (m_ssSamplesToDisplay.size() - 1) : (m_ssDisplayIdx - 1);
    }
  }

  
  public void showSamplesStepForward() {
    // Step forward to the next sample in ShowSamples mode when under manual user control
    m_ssAutoUpdate = false;
    m_ssDisplayIdx = m_ssSamplesToDisplay.isEmpty() ? 0 : (m_ssDisplayIdx+1) % m_ssSamplesToDisplay.size();
  }
  
  
  public void showSamplesDecrementNumSamples() {
    // Decrease the number of samples to be displayed in ShowSamples mode
    if (m_ssMaxSamplesToDisplay > 1) {
      m_ssMaxSamplesToDisplay--;
    }
  }
  
  
  public void showSamplesIncrementNumSamples() {
    // Increase the number of samples to be displayed in ShowSamples mode
    if ((m_ssMaxSamplesToDisplay < 100) && (m_ssMaxSamplesToDisplay < m_allSelectedSamples.size())) {
      m_ssMaxSamplesToDisplay++;
    }
  }
  
  
  public int getSampleHue(int sampleDisplayIdx) {
    // returns a Hue value between 0 and 255 to be used for displaying a specific sample in ShowSamples mode,
    // from the current palatte defined in m_ssSampleHues according to the index number passed in
    
    return m_ssSampleHues.get(sampleDisplayIdx % m_ssSampleHues.size());
  }

  
  public void mousePressed() {
    
    if (m_bShowHelpScreen) {
      hideHelpScreen();
      return;
    }
    
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      node.mousePressed();
    }
    for (BrushLink link : m_brushLinks) {
      link.mousePressed();
    }    
  }
 
 
  public void mouseReleased() {
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      node.mouseReleased();
    }      
  }
  
  
  public void mouseDragged() {
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      node.mouseDragged();
    }     
  }
  
  
  public void requestCausalLinkDelete() {
    // The user has pressed 'C' to delete a casual link.
    // Check whether the mouse pointer is currently over a causal link's control handle.
    // If so, delete that link, otherwise do nothing.
    
    for (CausalLink link : m_causalLinks) {
      if (link.mouseOver()) {
        m_causalLinks.remove(link);
        return;
      }
    }
  }
  
  
  public void requestToggleCausalLinkDir() {
    // The user has pressed 'D' to reverse the direction of a casual link.
    // Check whether the mouse pointer is currently over a causal link's control handle.
    // If so, toggle the link's direction, otherwise do nothing.
    
    for (CausalLink link : m_causalLinks) {
      if (link.mouseOver()) {
        link.toggleDirection();
        return;
      }
    }    
  }
  
  
  public void requestBrushLinkDelete() {
    // The user has pressed 'B' to delete an existing brush link.
    // This method checks if the mouse pointer is currently over a link's control handle.
    // If so, we delete that link. If the pointer is not over any brush link's handle,
    // do nothing.
    
    for (BrushLink link : m_brushLinks) {
      if (link.mouseOver()) {
        m_brushLinks.remove(link);
        return;
      }
    }    
  }
  
  
  public void requestBrushLinkCreate() {
    // The user has pressed 'L' to create a new brush link.
    // This method checks if the mouse pointer is currently over a node, and whether we have already
    // initiated a new brush link creation process (in which case this is the second node of the link
    // rather than the first).
    
    checkAllNodesSafe();
    for (Node node : m_allNodes) {
      if (node.mouseOver()) {
        // mouse is over this node, so let's investigate further...
        if (m_bNewBrushLinkUnderConstruction) {
          if (checkProposedBrushLinkValid(node)) {
            BrushLink link = new BrushLink(this, m_newBrushLinkNode1, node, 1.0f);
            m_brushLinks.add(link); 
          }
          else {
            // tried to create a link, but it is not valid. Indicate this to the user
            // TO DO... (at some point in the future)
          }
          m_newBrushLinkNode1.setBrushLinkUnderConstruction(false);
          m_bNewBrushLinkUnderConstruction = false;
        }
        else {
          // this is the first node specified for a new link
          m_bNewBrushLinkUnderConstruction = true;
          m_newBrushLinkNode1 = node;
          node.setBrushLinkUnderConstruction(true);
        }
        break;
      }
    }

  }
  
  
  public boolean checkProposedBrushLinkValid(Node node) {   
    // Perform the following checks on whether the proposed creation of a new
    // brush link between m_newBrushLinkNode1 and node is valid:
    // 1. Check that we're not trying to link a node to itself
    // 2. Check that both nodes currently have focus
    // 3. Check that the link doesn't already exist
    // 4. Check that the creation of the link does not create any circular links between nodes
    
    assert(m_bNewBrushLinkUnderConstruction);
        
    boolean valid = true;
    
    // Check 1
    if (node == m_newBrushLinkNode1) {
      return false;
    }
    
    // Check 2
    if (!(node.m_bHasFocus && m_newBrushLinkNode1.m_bHasFocus)) {
      return false;
    }

    // Check 3
    for (BrushLink link : m_brushLinks) {
      if ((link.m_node1 == node && link.m_node2 == m_newBrushLinkNode1) ||
          (link.m_node2 == node && link.m_node1 == m_newBrushLinkNode1)) {
        return false;
      }
    }
    
    // Check 4
    // Actually, we can get away without this at the moment, as we don't propagate linked
    // brushing in any case, so circular links do not cause a problem
    // TO DO... at some point in the future, if and when we implement propagation of linked brushing
    
    return valid;
  }
  
  
  public void adjustBrushLinkedNodes(Node node) {
    // This method is called when node's range selector bar has been dragged.
    // Look for any other nodes that are linked to node in m_brushLinks. If any are found, adjust
    // their range selector bars according to the nature of the link.
    
    for (BrushLink link : m_brushLinks) {
      if (link.enabled()) {
        if (link.m_node1 == node) {
          link.m_node2.adjustRangeSelector(node, link.m_strength);
        }
        else if (link.m_node2 == node) {
          link.m_node1.adjustRangeSelector(node, link.m_strength);
        }
      }
    }
  }
  
  
  public void toggleMeanMedian() {
    // switch between display of means and medians in the nodes
    m_showMedians = !m_showMedians;
  }
  
  
  public void toggleSSDisplayMode() {
    // for ShowSamples mode, switch between showing samples in random colors, or colored according to which
    // bin the sample appears in in the first focal node
    m_ssDisplayMode = (m_ssDisplayMode + 1) % 3;
    updateSelectedSampleList();
  }
  
  
  public void addMinimisedNode(Node node) {
    assert(!m_minimisedNodes.contains(node));
    m_minimisedNodes.add(node);
  }
  
  
  public void removeMinimisedNode(Node node) {
    assert(m_minimisedNodes.contains(node));
    m_minimisedNodes.remove(node);
  }
  
  
  public void drawMinimisedNodes() {
    // figure out position of each node, and setMinimisedPos before drawing it
    
    int x = m_minimisedInternodeGap;
    int y = height - m_minimisedInternodeGap;
    
    for (Node node : m_minimisedNodes) {
         
      if ((x > m_minInternodeGap) && (x + node.getMinimisedW() > width)) {
        y -= (node.getMinimisedH() + m_minimisedInternodeGap);
        x = m_minimisedInternodeGap;
      }
    
      node.setMinimisedPos(x, y - node.getMinimisedH());
      node.drawMinimised();
    
      x += (node.getMinimisedW() + m_minimisedInternodeGap);
      if (x > width) {
        y -= (node.getMinimisedH() + m_minimisedInternodeGap);
        x = m_minimisedInternodeGap;
      }
      
    }
  }
  
  
  public void toggleHelpScreen() {
    m_bShowHelpScreen = !m_bShowHelpScreen;
  }
  
  
  public void showHelpScreen() {
    m_bShowHelpScreen = true;
  }
  
  
  public void hideHelpScreen() {
    m_bShowHelpScreen = false;
  }


}

public abstract class Node {
  
  // identity
  int m_id;               // used to refer to this Node when we can't use a reference
  String m_name;          // human readable name of node
  
  // Widget appearance and position
  int m_nodeW;            // width of Node widget
  int m_nodeH;            // height of Node widget
  int m_referenceH;       // reference height of Node, used when rescaling
  int m_x;                // x position of top-left corner
  int m_y;                // y position of top-left corner
  int m_minimisedX;       // x position when node is minimised
  int m_minimisedY;       // y position when node is minimised
  int m_minimisedW = 100; // width when minimised
  int m_nodeZoom = 100;
    
  protected int m_mbH = 25;       // menu bar height
  protected int m_mbWidgetW = 25; // width of a widget in the menu bar
  protected int m_mbSpacer = 7;   // space used for placement of various items in menu bar
  protected int m_hgH;            // histogram height (DERIVED = m_nodeH - m_mbH - m_rsH - m_lbH)
  protected int m_hgHeadH = 12;   // histogram header height (for drawing bin base lines): this is part of m_hgH
  protected int m_hgFootH = 7;    // histogram footer height (for drawing bin base lines): this is part of m_hgH
  protected int m_rsH = 25;       // range selector height
  protected int m_lbH = 25;       // label bar height
  
  int m_mbBackgroundColor;
  int m_mbFocusColor;
  int m_mbRootColor;
  int m_mbInterColor;
  int m_mbLeafColor;
  int m_mbMinWidgetBackgroundColor;
  int m_mbMinWidgetForegroundColor;
  int m_hgBackgroundColor;
  int m_hgBaseColor;
  int m_rsBackgroundColor;
  int m_rsHandleColor;
  int m_rsFocalMeanValColor;
  int m_rsNonFocalMeanValColor;
  int m_rsHandlePressedColor;
  int m_rsSelectedRangeColor;
  int m_lbBackgroundColor;
  int m_lbForegroundColor;
  
  // Interaction
  boolean m_bNodeDragged;
  boolean m_bHasFocus;
  boolean m_bMinimised;
  boolean m_rsLeftHandlePressed;
  boolean m_rsRightHandlePressed;
  boolean m_rsBarPressed;
  boolean m_lbNodeResizeHandlePressed;
  int m_mousePressX;          // records x pos of where mouse was last pressed
  int m_mousePressY;          // records y pos of where mouse was last pressed
  int m_rsMousePressLLDeltaX; // records distance from m_mousePressX to left side of left handle when mouse last pressed
  int m_rsMousePressRRDeltaX; // records distance from m_mousePressX to right side of right handle when mouse last pressed
  int m_lbNodeResizeDeltaX;   // records distance from from mouse press to left edge of node when mouse pressed to resize node
  int m_lbNodeResizeDeltaY;   // records distance from from mouse press to bottom edge of node when mouse pressed to resize node
  
  // References to data associated with this Node
  Model m_model;      // reference to the associated Model
  int m_dataFileCol;  // data for this node is found in this column of the data file
                      // N.B. this is a 1-based index!
  int m_dataArrayCol; // data for this node is found in model.m_data.get(row).get(m_dataArrayCol)
                      // N.B. this is a 0-based index! (DERIVED = m_dataFileCol-1)
  
  // Histogram 
  int m_sMaxBins = 20;
  int m_hgNumBins;
  int[] m_hgBinSampleCounts;        // stores number of samples in each bin
  ArrayList<HistogramBin> m_hgBins;
  ArrayList<Number> m_hgBinMinVals; // stores the min limit of range stored in each bin
  int m_hgMinInterBinGap = 8;
  int m_hgMaxInterBinGap = 15;
  int m_hgDefaultMaxBinH;       // max allowed bin height (except for FullAutoHeightAdjust) (DERIVED = m_hgH-m_hgHeadH-m_hgFootH)
  int m_hgDefaultBinW = 8;      // standard bin width (for all but FullAreaConserved)
  int m_hgNumBrushes = 3;       // max number of different brushes/colors to use in the histogram bins
  int m_hgBaseStrokeWeight = 1; // width of base line below each bin
  
  // Range Selector
  int m_rsLow = 0;   // the (0-based) index in m_hgBins of the lowest selected histogram bin
  int m_rsHigh = 0;  // the (0-based) index in m_hgBins of the highest selected histogram bin
  
  // Links to causally connected Nodes
  ArrayList<Integer> m_parentIDs; // list of IDs (Node.m_id) of this node's causal parents 
  int m_role;                     // specifies whether this node is a Root (0), Intermediate (1) or Leaf node (2)
  
  // Information about BrushLinks associated with this node
  boolean m_brushLinkUnderConstruction; // flag to indicate if user is in process of constructing a new brush link on this node
  
  // Abstract classes to be specialised in subclasses
  public abstract int    getFullRange();
  public abstract int    getSelectedRange();
  public abstract void   initialiseHistogram();
  public abstract int    getHistogramIndex(Number num);
  public abstract Number getHistogramBinLowVal(int bin);
  public abstract Number getHistogramBinHighVal(int bin);

  
  Node(Model model, int id, String name, int filecol, int datacol, int role, ArrayList<Integer> parentIDs) {
    m_model = model;
    m_id = id;
    m_name = name;
    m_dataFileCol = filecol;
    m_dataArrayCol = datacol;
    m_role = constrain(role, 0, 2);
    m_parentIDs = parentIDs;
    m_nodeW = getDefaultW();
    m_nodeH = getDefaultH();
    m_referenceH = m_nodeH;
    m_hgNumBins = 10;
    m_hgBins = new ArrayList<HistogramBin>();
    m_hgH = m_nodeH - m_mbH - m_rsH - m_lbH;
    m_hgDefaultMaxBinH = m_hgH - m_hgHeadH - m_hgFootH; 
    m_mbBackgroundColor      = 0xffE0E0E0;
    m_mbFocusColor           = 0xffEE2222;
    colorMode(HSB, 360, 100, 100);
    m_mbRootColor            = color(120, 35, 80);
    m_mbInterColor           = color(170, 35, 80);
    m_mbLeafColor            = color(220, 35, 80);
    colorMode(RGB);
    m_mbMinWidgetBackgroundColor = 0xffCCCCCC;
    m_mbMinWidgetForegroundColor = 0xff101010;   
    m_hgBackgroundColor      = 0xffFFFFFF;
    m_hgBaseColor            = 0xff000000;
    m_rsBackgroundColor      = 0xff999999;
    m_rsHandleColor          = 0xffBBDDDD;
    m_rsFocalMeanValColor    = 0xff880000;
    m_rsNonFocalMeanValColor = 0xff000088;
    m_rsHandlePressedColor   = 0xffDDFFFF;
    m_rsSelectedRangeColor   = 0xff99BBBB;
    m_lbBackgroundColor      = 0xffE0E0E0;
    m_lbForegroundColor      = 0xff101010;
    m_bHasFocus = false;
    m_bMinimised = false;
    m_bNodeDragged = false;
    m_rsLeftHandlePressed = false;
    m_rsRightHandlePressed = false;
    m_rsBarPressed = false;
    m_lbNodeResizeHandlePressed = false;
    m_mousePressX = 0;
    m_mousePressY = 0;
    m_rsMousePressLLDeltaX = 0;
    m_rsMousePressRRDeltaX = 0;
    m_lbNodeResizeDeltaX = 0;
    m_lbNodeResizeDeltaY = 0;
    m_x = (int)random(0, width - m_nodeW);
    m_y = (int)random(0, height - m_nodeH);
    m_brushLinkUnderConstruction = false;
  }
  
  
  public int getDefaultW() {
    return m_model.m_nodeDefaultWidth;
  }
  
  
  public int getDefaultH() {
    return m_model.m_nodeDefaultHeight;
  }
  
  
  public void setH(int h, boolean resizeBins, boolean resetReferenceH) {
    // Set the height of the node, and also resize the histogram bins if requested.
    // A node has a concept of a reference height, which is used for calculating scale factors when
    // resizing histogram bins. This solves the problem of accumulating rounding errors in bin heights
    // after multiple resizes, which might otherwise occur.
     
    int nonHistH = m_mbH + m_rsH + m_lbH + m_hgHeadH + m_hgFootH; // combined height of everything except main histogram area
    int oldHistH = m_referenceH - nonHistH;
    int newHistH = h - nonHistH;
    float histSF = (float)newHistH / (float)oldHistH;
    
    m_nodeH = h;
    m_hgH = m_nodeH - m_mbH - m_rsH - m_lbH;
    
    if (resetReferenceH) {
      m_referenceH = h;
    }
    
    for (HistogramBin bin : m_hgBins) {
      bin.setY(m_hgH - m_hgFootH);
    }
 
    if (resizeBins) {
      for (HistogramBin bin : m_hgBins) {
        bin.scaleH(histSF, resetReferenceH);
      }      
    }
    
  }
  
  
  public void setW(int w) {
    // Attempt to change the node's width to the value specified.
    // The width of individual bins is not changed, so the minimum width allowed is
    // determined by the number of bins and their widths.

    if (w > 0) {
      // work out the inter-bin gap that the specified width would entail
      int gap = (int)((float)(w - m_hgNumBins*m_hgDefaultBinW) / (float)(m_hgNumBins+1));
      gap = constrain(gap, 0, 100); // now constrain the gap to a sensible number
      
      // adjust the x position of each bin
      int binx = gap;   
      for (HistogramBin bin : m_hgBins) {
        bin.setX(binx);
        binx += bin.m_w + gap;
      }
      
      // and finally set the node width according to how the bins are now spaced
      // (which should be equal to w unless the gap was constrained)
      m_nodeW = binx;
    }
  }
  
  
  public void setPosition(int x, int y) {
    m_x = x;
    m_y = y;
  }

  public void setX(int x) {
    m_x = x;
  }


  public void setY(int y) {
    m_y = y;
  }


  public int getH() {
    // returns the full height of the node
    return (m_mbH + m_hgH + m_rsH + m_lbH);
  }
  
  public int getW() {
    // returns the width of the node
    return m_nodeW;
  }
  
  
  public int getMinimisedH() {
    // returns the height of the node when minimised
    return m_mbH;
  }
  
  
  public int getMinimisedW() {
    return m_minimisedW;
  }  
  
  
  public int getCentreX() {
    // returns the x position of the centre of the node
    return (m_x + (m_nodeW / 2));
  }


  public int getCentreY() {
    // returns the y position of the centre of the node
    return (m_y + (getH() / 2));
  }
  
  
  public boolean hasFocus() {
    return m_bHasFocus;
  }

  
  public void initialiseHistogramCommon() {
    // This is a common helper method called in the initialiseHistogram methods of derived classes

    m_hgBinSampleCounts = new int[m_hgNumBins];
    m_hgBinMinVals = new ArrayList<Number>();
    
    // for each bin in the histogram, record the minimum value of its range
    ArrayList<ArrayList<Integer>> sampleIDs = new ArrayList<ArrayList<Integer>>();
    for (int i=0; i < m_hgNumBins; i++) {
      sampleIDs.add(new ArrayList<Integer>());
      m_hgBinMinVals.add(getHistogramBinLowVal(i));
    }
    
    // for each sample of data, add it to the appropriate bin
    int rowNum = 0;
    for (ArrayList<Number> row : m_model.m_data) {
      Number data = row.get(m_dataArrayCol); // from this data sample, get the column corresponding to this node
      int idx = getHistogramIndex(data);     // get the bin (index of m_hgBins) that this data value belongs to
      if (idx < 0) {
        println("Out of range number "+data.toString()+" encountered in column "+(m_dataArrayCol+1)+" of line "+(rowNum+2)+" of data file. Aborting.");
        exit();
      }
      //println(m_name+": rowNum="+rowNum+", m_dataArrayCol="+m_dataArrayCol+", data="+data+", idx="+idx);
      m_hgBinSampleCounts[idx]++;
      sampleIDs.get(idx).add(rowNum);
      rowNum++;
    }

    m_rsHigh = m_hgNumBins-1;
    
    // work out the appropriate spacing between bins in the histogram
    int gap = (int)((float)(m_nodeW - (m_hgNumBins * m_hgDefaultBinW)) / (float)(m_hgNumBins + 1)); 
    gap = constrain(gap, m_hgMinInterBinGap, m_hgMaxInterBinGap);
    int binx = gap;
    int maxH = 0;
    
    //println("Node "+m_name+": number of samples per bin:");
    // create a new HistogramBin object for each bin
    for (int i=0; i<m_hgNumBins; i++) {
      HistogramBin bin = new HistogramBin(this, m_hgBins.size(), m_hgBinSampleCounts[i], sampleIDs.get(i), binx, m_hgH - m_hgFootH);
      m_hgBins.add(bin);
      maxH = max(maxH, bin.getH());
      binx += bin.m_w + gap;
      //println("  bin "+(i+1)+": "+m_hgBinSampleCounts[i]);
    }
    
    // reset width of node according to space taken up by the bins
    m_nodeW = binx;
    
    // if the max bin height is greater than the height allowed by the node, increase height of node to fit
    int newH = m_mbH + m_rsH + m_lbH + m_hgHeadH + maxH + m_hgFootH;
    if (newH > m_nodeH) {
      setH(newH, false, true);    
    }
    
  }
  
  
  public int getMaxSamplesPerBin() {
    int max = 0;
    for (HistogramBin bin : m_hgBins) {
      if (bin.m_numSamples > max) {
        max = bin.m_numSamples;
      }
    }  
    return max;
  }


  public void draw(int nodeZoom) {
    
    if (m_bMinimised) {
      return;
    }
    
    m_nodeZoom = nodeZoom;
    
    pushMatrix();
    pushStyle();

    scale(((float)m_model.m_globalZoom)/100.0f);
    
    translate(m_x, m_y);
    
    drawMenuBar();
    drawHistogram();
    drawRangeSelector();
    drawLabelBar();
    
    if (m_brushLinkUnderConstruction) {
      // draw a border around the node to indicate that this has been selected as the first node of
      // a new brush link
      pushStyle();
      strokeWeight(2);
      stroke(0xffFF0000); // TO DO: define this (and weight) as class variables
      noFill();
      rect(0, 0, m_nodeW, m_nodeH);
      popStyle();
    }    
    
    popStyle();
    popMatrix();
  }

  
  public void drawMenuBar() {
    pushStyle();
    pushMatrix();
    
    // draw background
    fill(m_mbBackgroundColor);
    rect(0, 0, m_nodeW, m_mbH);
    
    // draw root/inter/leaf indication
    switch (m_role) {
      case 0: fill(m_mbRootColor);  break;
      case 1: fill(m_mbInterColor); break;
      case 2: fill(m_mbLeafColor);  break;
      default: fill(m_mbBackgroundColor);
    }
    rect(0, 0, m_nodeW, m_mbSpacer);
    
    // draw focus indicator
    if (m_bHasFocus) {
      stroke(m_mbFocusColor);
      fill(m_mbFocusColor);
      rect(m_mbSpacer-1, m_mbSpacer, m_mbWidgetW-(2*m_mbSpacer)+1, m_mbH-(2*m_mbSpacer)+1);
    }    
    
    // draw minimize widget
    fill(m_mbMinWidgetBackgroundColor);
    stroke(m_mbMinWidgetForegroundColor);
    rect(m_nodeW-m_mbWidgetW+m_mbSpacer, m_mbSpacer, m_mbWidgetW-(2*m_mbSpacer), m_mbH-(2*m_mbSpacer));
    strokeWeight(2);
    strokeCap(SQUARE);
    line(m_nodeW-m_mbWidgetW+m_mbSpacer, m_mbH-m_mbSpacer-1, m_nodeW-m_mbSpacer, m_mbH-m_mbSpacer-1);
    
    popMatrix();
    popStyle();
  }

  
  public void drawHistogram() {
    pushMatrix();
    translate(0, m_mbH);
    fill(m_hgBackgroundColor);
    rect(0, 0, m_nodeW, m_hgH);
    
    if (m_hgBins != null) {
      for (HistogramBin bin : m_hgBins) {
        bin.draw();
      }
    }
    
    // draw labels for min and max selected values
    String lowVal = getLowValueString();
    String highVal = getHighValueString();
    int lowX = m_hgBins.get(m_rsLow).getLX();
    int highX = m_hgBins.get(m_rsHigh).getRX();
    int textY = (int)((float)m_hgHeadH * 0.8f);
    textFont(m_model.m_smallFont, 11);
    fill(m_lbForegroundColor);
    textAlign(LEFT);
    text(lowVal, lowX, textY);
    if (m_rsLow != m_rsHigh) {
      textAlign(RIGHT);
      text(highVal, highX, textY);
    }    
    
    popMatrix();
  }

  
  public void drawRangeSelector() {
    
    pushStyle();
    pushMatrix();
    
    translate(0, m_mbH+m_hgH);
    fill(m_rsBackgroundColor);
    rect(0, 0, m_nodeW, m_rsH);

    if (m_hgBins != null) {
      int llx = m_hgBins.get(m_rsLow).getLX();
      int lrx = m_hgBins.get(m_rsLow).getRX();
      int hlx = m_hgBins.get(m_rsHigh).getLX();
      int hrx = m_hgBins.get(m_rsHigh).getRX();
      // draw range bar
      fill(m_rsSelectedRangeColor);
      rect(lrx, 0, hlx-lrx, m_rsH);
      // draw left handle
      fill(m_rsLeftHandlePressed ? m_rsHandlePressedColor : m_rsHandleColor);
      rect(llx, 0, lrx-llx, m_rsH);
      //
      if (m_rsLow == m_rsHigh) {
        // left and right handles are in the same place, so
        // draw partitions on handle for expand right, drag and expand left
        stroke(m_rsBackgroundColor);
        line(llx, m_rsH/3, lrx, m_rsH/3);
        line(llx, (2*m_rsH)/3, lrx, (2*m_rsH)/3);
        noStroke();
      }
      else {
        // draw right handle
        fill(m_rsRightHandlePressed ? m_rsHandlePressedColor : m_rsHandleColor);
        rect(hlx, 0, hrx-hlx, m_rsH);
      } 
      
      // draw the mean or median value of the selected range
      if (m_rsLow != m_rsHigh) {
        float mv = (m_model.m_showMedians) ? getMedianSelectedValue(!m_bHasFocus) : getMeanSelectedValue(!m_bHasFocus);
          // mv is the mean/median value of samples that lie in the currently selected range
        float lowv  = getHistogramBinLowVal(m_rsLow).floatValue();   // low boundary of lowest bin in selected range
        float highv = getHistogramBinHighVal(m_rsHigh).floatValue(); // high boundary of highest bin in selected range
        
        if (mv < 0.0f) { // getMean and getMedian return a negative number if there are no selected samples
          mv = lowv;
        }
        
        int lmx = (llx+lrx)/2;
        int hmx = (hlx+hrx)/2;
        int mx = lmx + (int)(((mv-lowv)*(float)(hmx-lmx))/(highv-lowv));
        int circ = (int)(0.5f * (float)m_rsH);
        fill(m_bHasFocus ? m_rsFocalMeanValColor : m_rsNonFocalMeanValColor);
        ellipse(mx, m_rsH/2, circ, circ);
        
        // print a mu character in the middle of the circle representing the mean value
        textFont(m_model.m_smallFont, 11);
        fill(0xFFFFFFFF);
        textAlign(CENTER);
        if (m_model.m_showMedians) {
          text(m_model.m_strXTilde, mx, (m_rsH*11)/16);
        }
        else {
          text(m_model.m_strMu, mx, (m_rsH*9)/16);
        }
      }
      
    }
    
    popMatrix();
    popStyle();
  }
  
  
  public void drawLabelBar() {
    pushStyle();
    pushMatrix();
    translate(0, m_mbH+m_hgH+m_rsH);
    // draw background
    fill(m_lbBackgroundColor);
    rect(0, 0, m_nodeW, m_lbH);
    // write node name    
    textFont(m_model.m_mediumFont, 16);
    fill(m_lbForegroundColor);
    textAlign(CENTER);
    text(m_name, m_nodeW/2, m_lbH-8);
    // draw node resize graphic
    stroke(m_rsBackgroundColor); // a bit of color sharing going on here
    strokeCap(PROJECT);
    for (int l=0; l<4; l++) {
      line( m_nodeW-(((4-l)*m_lbH)/8), m_lbH, m_nodeW, 1+(((4+l)*m_lbH)/8) );
    }
    // and finally revert to previous state
    popMatrix();
    popStyle();
  }
  
  
  public void setFullRange() {
    // set the high and low range selectors to their extreme values
    m_rsLow = 0;
    m_rsHigh = m_hgNumBins-1;
  }
  
  
  public boolean mouseOver() {
    // Returns true if the mouse pointer is currently over this node, otherwise false
    
    if (m_bMinimised) {
      return (mouseX >= m_minimisedX &&
              mouseX < m_minimisedX + m_minimisedW &&
              mouseY >= m_minimisedY &&
              mouseY <= m_minimisedY + m_mbH);      
    }
    else {
      return (scaledMouseX() >= m_x &&
              scaledMouseX() < m_x + m_nodeW &&
              scaledMouseY() >= m_y &&
              scaledMouseY() <= m_y + m_nodeH);
    }
  }
  
  
  public void mousePressed() {
    
    if (m_bMinimised) {
      mousePressedMinimised();
      return;
    }
    
    m_mousePressX = scaledMouseX();
    m_mousePressY = scaledMouseY();
    
    if (mouseOver()) {
      // the mouse has been pressed within this node, so figure out what we need to do about it!
      
      if (scaledMouseY() < m_y + m_mbH) {
        ///////////// MOUSE IS IN THE MENU BAR AREA ///////////////////////////////////////////////
        if (scaledMouseX() > m_x + m_nodeW - m_mbWidgetW) {
          minimise();
        }
        else {
          m_bNodeDragged = true;
        }        
      }
      else if (scaledMouseY() >= m_y + m_mbH && scaledMouseY() < m_y + m_mbH + m_hgH) {
        ///////////// MOUSE IS IN THE HISTOGRAM AREA ///////////////////////////////////////////////
        
        switch (m_model.m_interactionMode) {
          case SingleNodeBrushing: {
            m_model.setSingleFocus(m_id);
            m_model.brushAllNodesOnOneSelection(this);          
            break;
          }
          case MultiNodeBrushing: {
            m_model.toggleMultiFocus(m_id);
            m_model.brushAllNodesOnMultiSelection();          
            break;
          }
          case ColourBins: {
            m_model.setSingleFocus(m_id);
            m_model.brushAllNodesOnOneSelection(this);              
            m_model.updateSelectedSampleList();            
            break;
          }
          case ShowSamples: {
            //m_model.toggleMultiFocus(m_id);
            //m_model.brushAllNodesOnMultiSelection();
            m_model.setSingleFocus(m_id);
            m_model.brushAllNodesOnOneSelection(this);              
            m_model.updateSelectedSampleList();
            break;
          }
          default: {
            println("Unexpected interaction mode in Node.mousePressed()!");
          }
        }
      }
      else if (scaledMouseY() >= m_y + m_mbH + m_hgH && scaledMouseY() <= m_y + m_mbH + m_hgH + m_rsH) {
        ///////////// MOUSE IS IN THE RANGE SELECTOR AREA //////////////////////////////////////////
        
        if (rangeSelectorActive()) {
          int llx = m_hgBins.get(m_rsLow).getLX();
          int lrx = m_hgBins.get(m_rsLow).getRX();
          int hlx = m_hgBins.get(m_rsHigh).getLX();
          int hrx = m_hgBins.get(m_rsHigh).getRX();
          
          if (m_rsLow == m_rsHigh) {
            // first deal with special case where both range selectors are in the same position
            if (scaledMouseX() >= m_x + llx && scaledMouseX() <= m_x + lrx) {
              if (scaledMouseY() <= (m_y + m_mbH + m_hgH + (m_rsH/3))) {
                // is top third of handle pressed, call it a right handle press
                m_rsRightHandlePressed = true;
              }
              else if (scaledMouseY() >= (m_y + m_mbH + m_hgH + ((2*m_rsH)/3))) {
                // else if bottom third pressed, call it a left handle press
                m_rsLeftHandlePressed = true;
              }
              else {
                // else handle pressed in middle, so treat it as a press of the rs bar for dragging
                m_rsBarPressed = true;
                m_rsMousePressLLDeltaX = m_mousePressX - (m_x + llx);
                m_rsMousePressRRDeltaX = (m_x + hrx) - m_mousePressX;
              }
            }
          }
          else {
            if (scaledMouseX() >= m_x + llx && scaledMouseX() <= m_x + lrx) {
              // left handle pressed
              m_rsLeftHandlePressed = true;
            }
            else if (scaledMouseX() >= m_x + hlx && scaledMouseX() <= m_x + hrx) {
              // right handle pressed
              m_rsRightHandlePressed = true;
            }
            else if (scaledMouseX() > m_x + lrx && scaledMouseX() < m_x + hlx) {
              // bin between the handles pressed
              m_rsBarPressed = true;
              m_rsMousePressLLDeltaX = m_mousePressX - (m_x + llx);
              m_rsMousePressRRDeltaX = (m_x + hrx) - m_mousePressX;
            }
          }
        }
        
      }
      else if (scaledMouseY() >= m_y + m_mbH + m_hgH + m_rsH) {
        ///////////// MOUSE IS IN THE LABEL BAR AREA ///////////////////////////////////////////////
        
        // calculate whether mouse is over the node resize button, and act accordingly
        int rl = m_lbH/2; // dimension of the resize button
        int dx = scaledMouseX() - (m_x + m_nodeW - rl);
        int dy = scaledMouseY() - (m_y + m_mbH + m_hgH + m_rsH + rl);
        if ((dx >= 0) && (dy >= 0) && ((rl-dx) <= dy)) {
          m_lbNodeResizeHandlePressed = true;
          m_lbNodeResizeDeltaX =  (m_x + m_nodeW) - scaledMouseX();
          m_lbNodeResizeDeltaY = (m_y + m_nodeH) - scaledMouseY();
        }
        else {
          m_bNodeDragged = true;
        }
      }
    }
  }


  public void mouseReleased() {
    m_bNodeDragged = false;
    m_rsLeftHandlePressed = false;
    m_rsRightHandlePressed = false;
    m_rsBarPressed = false;
    m_lbNodeResizeHandlePressed = false;
  }

  
  public void mouseDragged() {
    
    if (m_bMinimised) {
      return;
    }
    
    int rsLowOld = m_rsLow;
    int rsHighOld = m_rsHigh;
    
    if (m_bNodeDragged) {
      ///////////// WHOLE NODE DRAGGED /////////////////////////////////////////////////
      m_x += (scaledMouseX() - scaledPMouseX());
      m_y += (scaledMouseY() - scaledPMouseY());
      constrain(m_x, 0, width - m_nodeW);
      constrain(m_y, 0, height - m_nodeH);      
    }
    else if (m_rsLeftHandlePressed) {
      ///////////// RANGE SELECTOR LEFT HANDLE DRAGGED /////////////////////////////////
      int llx = m_hgBins.get(m_rsLow).getLX();
      int rrx = m_hgBins.get(m_rsLow).getRX();    
      if (scaledMouseX() < m_x) {
        m_rsLow = 0;
      }
      else if ((scaledMouseX() < m_x + llx) && (m_rsLow > 0)) {
        for (int i = m_rsLow-1; i >= 0; i--) {
          if (scaledMouseX() < m_x + m_hgBins.get(i).getRX()) {
            m_rsLow = i;
          }
          else {
            break;
          }
        } 
      }
      else if ((scaledMouseX() > m_x + rrx) && (m_rsLow < m_rsHigh)) {
        for (int i = m_rsLow+1; i <= m_rsHigh; i++) {
          if (scaledMouseX() > m_x + m_hgBins.get(i).getLX()) {
            m_rsLow = i;
          }
          else {
            break;
          }
        } 
      }      
    }
    else if (m_rsRightHandlePressed) {
      ///////////// RANGE SELECTOR RIGHT HANDLE DRAGGED ////////////////////////////////
      int llx = m_hgBins.get(m_rsHigh).getLX();
      int rrx = m_hgBins.get(m_rsHigh).getRX();    
      if (scaledMouseX() < m_x) {
        m_rsHigh = m_rsLow;
      }
      else if ((scaledMouseX() < m_x + llx) && (m_rsHigh > m_rsLow)) {
        for (int i = m_rsHigh-1; i >= m_rsLow; i--) {
          if (scaledMouseX() < m_x + m_hgBins.get(i).getRX()) {
            m_rsHigh = i;
          }
          else {
            break;
          }
        } 
      }
      else if ((scaledMouseX() > m_x + rrx) && (m_rsHigh < m_hgNumBins-1)) {
        for (int i = m_rsHigh+1; i <= m_hgNumBins-1; i++) {
          if (scaledMouseX() > m_x + m_hgBins.get(i).getLX()) {
            m_rsHigh = i;
          }
          else {
            break;
          }
        } 
      }
    }
    else if (m_rsBarPressed) {
      ///////////// RANGE SELECTOR BAR BETWEEN HANDLES DRAGGED /////////////////////////
      int llx = m_hgBins.get(m_rsLow).getLX();
      int rrx = m_hgBins.get(m_rsLow).getRX(); 
      if (scaledMouseX() < m_x) {
        // moving to extreme left
        m_rsHigh -= m_rsLow;
        m_rsLow = 0;
      }
      else if ((scaledMouseX() < scaledPMouseX()) && (m_rsLow > 0)) {
        // moving left
        for (int i = m_rsLow; i >= 0; i--) {
          if ((scaledMouseX() - m_rsMousePressLLDeltaX) < (m_x + m_hgBins.get(i).getRX())) {
            m_rsHigh -= (m_rsLow-i);
            m_rsLow = i;
          }
          else {
            break;
          }
        } 
      }
      else if ((scaledMouseX() > scaledPMouseX()) && (m_rsHigh < m_hgNumBins-1)) {
        // moving right
        for (int i = m_rsHigh+1; i <= m_hgNumBins-1; i++) {
          if (scaledMouseX() + m_rsMousePressRRDeltaX > m_x + m_hgBins.get(i).getLX()) {
            m_rsLow += (i-m_rsHigh);
            m_rsHigh = i;
          }
          else {
            break;
          }
        }         
      }
      // if this node's range selector bar has been dragged, we now need to move the range selector bars
      // of any nodes that are currently linked to this one
      m_model.adjustBrushLinkedNodes(this);
    }
    else if (m_lbNodeResizeHandlePressed) {
      ///////////// NODE RESIZE HANDLE DRAGGED /////////////////////////
      
      if ((scaledMouseX() > m_x + 0) && (scaledMouseY() > m_y + m_mbH + m_rsH + m_hgHeadH + m_hgFootH + m_lbH )) {
        setH(scaledMouseY() - m_y + m_lbNodeResizeDeltaY, true, false);
        setW(scaledMouseX() - m_x + m_lbNodeResizeDeltaX);
      }  
      
    }
    
    if (m_rsLow != rsLowOld || m_rsHigh != rsHighOld) {
      // If either the low or high range selector handles have moved, we now need to take some further action!
      switch(m_model.m_interactionMode) {
        case SingleNodeBrushing: {
          m_model.brushAllNodesOnOneSelection(this);
          break;
        }
        case MultiNodeBrushing: {
          m_model.brushAllNodesOnMultiSelection();
          break;
        }
        case ColourBins: {
          m_model.brushAllNodesOnOneSelection(this);
          m_model.updateSelectedSampleList();          
          break;
        }
        case ShowSamples: {
          //m_model.brushAllNodesOnMultiSelection();
          m_model.brushAllNodesOnOneSelection(this);
          m_model.updateSelectedSampleList();
          break;
        }
        default: {
          println("In Node.mouseDragged(): I got somewhere I shouldn't be!");
          exit();
        }
      }
    }
  }
  
  
  public ArrayList<Integer> getSelectedSampleIDs() {
    // return a list of sampleIDs of all samples within the selected range
    ArrayList<Integer> list = new ArrayList<Integer>();
    for (int i=m_rsLow; i<=m_rsHigh; i++) {
      list.addAll(m_hgBins.get(i).m_sampleIDs);
    }
    return list;
  }
  
  
  public void brushSamples(ArrayList<Integer> samples) {
    // cycle over all bins, brushing the appropriate number in each bin according to how many
    // samples in that bin are in the specified list of samples
    for (HistogramBin bin : m_hgBins) {
      bin.brushSamples(samples);
    }
  }


  public void brushSampleAdd(int sampleID, int numMisses) {
    // cycle over all bins, looking for the one that contains the specified sample. When the
    // matching bin is found, add the sample to that bin's list of brushed samples according
    // to the specied numMisses
    
    for (HistogramBin bin : m_hgBins) {
      if (bin.brushSampleAdd(sampleID, numMisses)) {
        break;
      }
    }
  } 
  
  
  public boolean sampleSelected(int sampleID) {
    // Returns true if the specified sample is in one of the currently selected range of bins
    boolean selected = false;
    for (int i=m_rsLow; ((i<=m_rsHigh) && (!selected)); i++) {
      selected = m_hgBins.get(i).sampleInBin(sampleID);
    }    
    return selected;
  }
  
  
  public void resetBrushing() {
    for (HistogramBin bin : m_hgBins) {
      bin.resetBrushing();
    }
  }
  
  
  public boolean rangeSelectorActive() {
    switch (m_model.m_interactionMode) {
      
      case SingleNodeBrushing:
        return m_bHasFocus;
      case MultiNodeBrushing:
        return m_bHasFocus;
      case ColourBins:
        return m_bHasFocus;
      case ShowSamples:
        return m_bHasFocus;
      default:
        return true;
    }
  }
  
  
  public String getLowValueString() {
    // Fetch the low bound of the lowest bin in the currently selected range
    // and return it as a string.
    // This copes with the fact that Node instances are either DiscreteNode or ContinuousNode
    // objects and the values will be ints or floats respectively. If the value is a float,
    // the string representation is limited to one decimal place.
    
    String val = new String(getHistogramBinLowVal(m_rsLow).toString());
    int dp = val.indexOf('.');
    if (dp >= 0) {
      val = val.substring(0, dp+2);
    }
    return val;
  }
  
  
  public String getHighValueString() {
    // Fetch the high bound of the highest bin in the currently selected range
    // and return it as a string.
    // This copes with the fact that Node instances are either DiscreteNode or ContinuousNode
    // objects and the values will be ints or floats respectively. If the value is a float,
    // the string representation is limited to one decimal place.
    
    String val = new String(getHistogramBinHighVal(m_rsHigh).toString());
    int dp = val.indexOf('.');
    if (dp >= 0) {
      val = val.substring(0, dp+2);
    }
    return val;
  } 
  
  
  public float getMeanSelectedValue() {
    // Calculate the mean value of samples in selected bins
    return getMeanSelectedValue(false);
  }
  
    
  public float getMeanSelectedValue(boolean brushedOnly) {
    // Calculate the mean value of samples in selected bins.
    // If brushedOnly==true, only count brushed samples (with numMiss==0), otherwise count all samples.
    // If there are no relevant samples, return -1.0.
    
    float total = 0;
    int numSamples = 0;
    for (int b = m_rsLow; b <= m_rsHigh; b++) {
      total += m_hgBins.get(b).getTotalValues(brushedOnly);    
      numSamples += (brushedOnly ? m_hgBins.get(b).numBrushed() : m_hgBins.get(b).numSamples());
    }
    
    if (numSamples == 0) {
      return -1.0f;
    }

    return total / (float)numSamples;
  }
  
  
  public float getMedianSelectedValue() {
    // Calculate the median value of samples in selected bins
    return getMedianSelectedValue(false);
  }
  
  
  public float getMedianSelectedValue(boolean brushedOnly) {
    // Calculate the median value of samples in selected bins.
    // If brushedOnly==true, only count brushed samples (with numMiss==0), otherwise count all samples.
    // If there are no relevant samples, return -1.0.
    
    ArrayList<Float> selectedValues = new ArrayList<Float>();

    for (int b = m_rsLow; b <= m_rsHigh; b++) {
      selectedValues.addAll(m_hgBins.get(b).getSelectedValues(brushedOnly));
    }
    
    if (selectedValues.isEmpty()) {
      return -1.0f;
    }
    
    Collections.sort(selectedValues);
    
    if (selectedValues.size() % 2 == 1) {
      return selectedValues.get((selectedValues.size()+1)/2-1);
    }
    else
    {
      float lower = selectedValues.get(selectedValues.size()/2-1);
      float upper = selectedValues.get(selectedValues.size()/2);
      return ((lower + upper) / 2.0f);
    }
  }
  
  
  public int scaledMouseX() {
    return (int)((float)(mouseX * 100.0f) / m_model.m_globalZoom);
  }

  
  public int scaledMouseY() {
    return (int)((float)(mouseY * 100.0f) / m_model.m_globalZoom);
  }


  public int scaledPMouseX() {
    return (int)((float)(pmouseX * 100.0f) / m_model.m_globalZoom);
  }
  
  
  public int scaledPMouseY() {
    return (int)((float)(pmouseY * 100.0f) / m_model.m_globalZoom);
  }
  
  
  public void setBrushLinkUnderConstruction(boolean flag) {
    m_brushLinkUnderConstruction = flag;
  }
  
  
  public void adjustRangeSelector(Node node, float strength) {
    // Adjust this node's range selector, based upon the position of the range selector of the other
    // node passed in to this method.
    int oMax = node.m_hgNumBins - 1;
    int oLow = node.m_rsLow;
    int oHigh = node.m_rsHigh;
    float otherRSPosFrac = ((float)(oLow + oHigh) / (float)(2 * oMax));  

    float thisCurrentRSPos = (float)m_rsLow + (((float)(m_rsHigh - m_rsLow)) / 2.0f);
    
    int thisMax = m_hgNumBins - 1;
    float thisTargetRSPos;
   
    if (strength >= 0.0f) {
      thisTargetRSPos = strength * otherRSPosFrac * (float)thisMax;
    }
    else {
      thisTargetRSPos = (1.0f + (strength * otherRSPosFrac)) * (float)thisMax;
    }
    
    int delta = round(thisTargetRSPos - thisCurrentRSPos);
    
    m_rsLow += delta;
    m_rsHigh += delta;
    
    if (m_rsLow < 0) {
      int d = -m_rsLow;
      m_rsLow += d;
      m_rsHigh += d;
    }
    else if (m_rsHigh >= m_hgNumBins) {
      int d = m_rsHigh + 1 - m_hgNumBins;
      m_rsLow -= d;
      m_rsHigh -= d;
    }
  }
  
  
  public ArrayList<Integer> getSamplesInRange() {
    // returns a list of IDs of all samples within the currently selected range of this node
    ArrayList<Integer> list = new ArrayList<Integer>();
    for (int b = m_rsLow; b <= m_rsHigh; b++) {
      list.addAll( m_hgBins.get(b).m_sampleIDs );
    }
    return list;
  }
  
  
  public void matchSampleBinsToColors(ArrayList<Integer> sampleIDs, ArrayList<Integer> hues) {
    // Takes a list of sampleIDs as input, and a list of hues to be returned as output. This list of
    // hues is first emptied. Then, go through each sample, and add an entry at the corresponding
    // position in the hue list according to the bin in which that sample is found.
    
    // First, empty the list of hues to start with a clean slate
    hues.clear();
    
    // Now create a hue for each bin in the histogram
    ArrayList<Integer> binHues = new ArrayList<Integer>(m_hgNumBins);
    for (int i=0; i<m_hgNumBins; i++) {
      binHues.add((i*255)/m_hgNumBins);
    }
    
    if (m_model.m_interactionMode == InteractionMode.ShowSamples && m_model.m_ssDisplayMode == 1) {
      Collections.shuffle(binHues);  // shuffle the hues so adjacent bins have very different hues
    }
    
    // Finally, for each sample in the list, determine which bin it belongs to and assign
    // a hue accordingly
    for (Integer sID : sampleIDs) {
      int binIdx = findSample(sID);
      
      assert(binIdx >= 0);
      if (binIdx < 0) {
        binIdx = 0;
      }
      
      hues.add(binHues.get(binIdx));
    }
  }
  
  
  public int findSample(int sampleID) {
    // Find which bin the specified sample belongs to, and return the index of the bin. If not found, return -1.
    for (HistogramBin bin : m_hgBins) {
      if (bin.sampleInBin(sampleID)) {
        return bin.m_idx;
      }
    }
    return -1;
  }
  
  
  public void minimise() {
    assert(!m_bMinimised);
    m_bMinimised = true;
    m_bHasFocus = false;
    m_minimisedX = m_x; // this will be reset by the model when it draws the minimised nodes
    m_minimisedY = m_y; // this will be reset by the model when it draws the minimised nodes
    m_model.addMinimisedNode(this);
  }
  
  
  public void maximise() {
    assert(m_bMinimised);
    m_bMinimised = false;
    m_model.removeMinimisedNode(this);
  }
  
  
  public void setMinimisedPos(int x, int y) {
    m_minimisedX = x;
    m_minimisedY = y;
  }
  
  
  public boolean minimised() {
    return m_bMinimised;
  }
  
  
  public void drawMinimised() {
    
    pushStyle();
    pushMatrix();
    
    translate(m_minimisedX, m_minimisedY);
    
    // draw background
    fill(m_mbBackgroundColor);
    rect(0, 0, m_minimisedW, m_mbH);
    
    // draw root/inter/leaf indication
    switch (m_role) {
      case 0: fill(m_mbRootColor);  break;
      case 1: fill(m_mbInterColor); break;
      case 2: fill(m_mbLeafColor);  break;
      default: fill(m_mbBackgroundColor);
    }
    rect(0, 0, m_minimisedW, m_mbSpacer);  
    
    // draw maximise widget
    fill(m_mbMinWidgetBackgroundColor);
    stroke(m_mbMinWidgetForegroundColor);
    rect(m_minimisedW-m_mbWidgetW+m_mbSpacer, m_mbSpacer, m_mbWidgetW-(2*m_mbSpacer), m_mbH-(2*m_mbSpacer));
    strokeWeight(2);
    strokeCap(SQUARE);
    line(m_minimisedW-m_mbWidgetW+m_mbSpacer, m_mbSpacer+1, m_minimisedW-m_mbSpacer, m_mbSpacer+1);
    
    // write node name    
    textFont(m_model.m_smallFont, 11);
    fill(m_lbForegroundColor);
    textAlign(CENTER);
    text(m_name.substring(0,min(m_name.length(),15)), m_minimisedW/2, m_mbH-6);
    
    popMatrix();
    popStyle();    
    
  }
  
  
  public void mousePressedMinimised() {
    // look for mouse presses on a minimised node and react accordingly

    m_mousePressX = scaledMouseX();
    m_mousePressY = scaledMouseY();
        
    if (mouseOver()) {
      // the mouse has been pressed within this node, so figure out what we need to do about it!
      if (mouseX > m_minimisedX + m_minimisedW - m_mbWidgetW) {
        maximise();
      }
    }
    
  }

  
}
  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "omnigram" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
